#open "palm";;
#open "events";;

let vshift = 15;;

(* Dimensions du damier *)
let bits_width = 15 ;;
let bits_height =  14 ;;

(* Nombre de motifs differents possibles *)
let pattern_number = 4 ;;

(* Surface marquee lors d'un parcours - Side effect *)
let marked_surface = ref 0 ;;

(* Nombre de cases restant a eliminer *)
let remaining_cases = ref 0 ;;

(* Couleur de la case designee pour le coup courant - Side effect *)
let current_color = ref 0 ;;

(* Tableau du jeu (0,0)=(haut, gauche) du damier *)
let bits = vect__make_matrix bits_width bits_height 0 ;;

(* Patterns: *)
let white = "\000\000\000\000\000\000\000\000";;
let black = "\255\255\255\255\255\255\255\255";;
let gray1 = "\170\085\170\085\170\085\170\085";;
let barv =  "\170\170\170\170\170\170\170\170";;
let barh =  "\255\000\255\000\255\000\255\000";;

(* Tableau qui fait correspondre une couleur a un numero de case *)
let color_mapping = [| white; black; gray1; barv; barh |];;

(* Taille d'un carre representant une case *)
let img_size = 10 ;;

(* Score courant *)
let score = ref 0 ;;

(* Flag disant si on doit generer aleatoirement le tableau ou non *)
let current_seed = ref 0 ;;


(* ************************* *)
(* Creation d'un nouveau jeu *)
(* ************************* *)
let init_bits () =
 (* Allocation du damier *)
 for x=0 to bits_width-1 do
  for y=0 to bits_height-1 do
   bits.(x).(y) <- (random__int (pattern_number)) + 1
  done
 done ;
 (* Nombre de cases restant a effacer *)
 remaining_cases := bits_width * bits_height
;;



(* ********************************************************************** *)
(* Marquage des cases adjacentes a une case, pourvu que la couleur de ces *)
(* voisins soit la meme que la couleur de la case originalement designee. *)
(* ********************************************************************** *)
let rec mark_bits x y =
 (* Si la case n'est pas marquee, on bosse *)
 if bits.(x).(y) land 0x80 = 0 then
   begin
   (* Si la case est de la couleur qui va bien, on bosse *)
   if bits.(x).(y) land 0x7F = !current_color then
    begin
    (* Marquer la case *)
    bits.(x).(y) <- bits.(x).(y) lor 0x80 ;
    (* Il faudra verifier a la fin que la surface *)
    (* reellement marquee est superieure a 1      *)
    incr marked_surface ;
    (* Explorer les voisins si c'est possible *)
    if x-1 >= 0 then mark_bits (x-1) y ;
    if x+1 < bits_width then mark_bits (x+1) y ;
    if y-1 >= 0 then mark_bits x (y-1) ;
    if y+1 < bits_height then mark_bits x (y+1)
    end
   end
;;



(* ******************************************************** *)
(* Effacement des cases marquees. Compactage si necessaire. *)
(* ******************************************************** *)
let delete_bits () =
 for x = 0 to bits_width-1 do
  (* On traite les colonnes de gauche a droite *)
  let next_room = ref (bits_height-1)  in
  for y = bits_height-1 downto 0 do
   (* On traite chaque ligne du bas vers le haut *)
   if bits.(x).(y) land 0x80 <> 0 then
    begin
    (* La case est marquee, donc on la "vide" *)
    bits.(x).(y) <- 0 ;
    decr remaining_cases
    end
   else
    begin
    (* La case est a 0 (donc "vide"), donc par construction, *)
    (* les cases du dessus le sont aussi                     *)
    if bits.(x).(y) = 0 then ()
    else
     begin
     (* La case n'est pas vide et n'est pas marquee *)
     let tmp = bits.(x).(y) in
     bits.(x).(y) <- 0 ;
     bits.(x).(!next_room) <- tmp ;
     decr next_room ;
     end
    end
  done
 done ;
 (* Maintenant, il faut scaner (de droite a gauche) les colonnes *)
 (* vides et les rapprocher a gauche pour faire du compactage    *)
 for x = bits_width-1 downto 0 do
  (* Si la case du bas est a 0, alors c'est que la colonne est vide *)
  if bits.(x).(bits_height-1) = 0 then
   begin
   (* x = indice de la colonne vide *)
   (* On rapproche toutes les colonnes se trouvant *)
   (* a droite en scanant de gauche a droite       *)
   for x'= x+1 to bits_width-1 do
    if x' < bits_width then
     begin
     (* La colonne x' va dans colonne x'-1 et on *)
     (* efface le contenu de la colonne x'.      *)
     for y=0 to bits_height-1 do
      bits.(x'-1).(y) <- bits.(x').(y) ;
      bits.(x').(y) <- 0
     done
     end
   done
   end
 done
;;



(* ************************************************** *)
(* Raffraichissement total de l'ecran                 *)
(* Version pour un ecran organise (0,0)=(bottom,left) *)
(* ************************************************** *)
let refresh_whole () =
  for y=0 to bits_height-1 do
    for x=0 to bits_width-1 do
      let color = bits.(x).(y) in
      win__set_pattern color_mapping.(color) ;
      	win__fill_rectangle 
	  ({ x = x*img_size; y = (y*img_size) + vshift},
	   { x = img_size; y = img_size })
	  0
  done
 done
;;

(* *********************************************** *)
(* Verification de s'il reste des coups possibles  *)
(* On parse la tableau de gauche a droite et de    *)
(* bas en haut pour voir s'il reste au moins 2     *)
(* cases adjacentes en haut ou a droite de la meme *)
(* couleur que nous.                               *)
(* *********************************************** *)
let check_game_over () =
 try
  begin
  for x = 0 to bits_width-2 do
   for y = bits_height-1 downto 1 do
    if bits.(x).(y) <> 0 then
     if bits.(x).(y) = bits.(x).(y-1) || bits.(x).(y) = bits.(x+1).(y)
      then raise Not_found
   done
  done ;
  (* Si on arrive ici c'est qu'on n'a rien trouve et donc le jeu est fini *)
  true
  end
 with
  | Not_found -> false
;;


(* ************************** *)
(* Affichage du score courant *)
(* ************************** *)
let refresh_score form =
  let s = "Score : " ^ string_of_int !score in
  frm__copy_title form s
;;

let busy form =
  let s = "Score : " ^ string_of_int !score ^ "..."in
  frm__copy_title form s
;;


(* *************************************** *)
(* Gestion d'un coup demande par ke joueur *)
(* *************************************** *)
let handle_turn form case_x case_y =
  (* On verifie qu'on est bien dans l'espace de jeu et que la case *)
  (* designee n'est pas une case vide *)
  if case_x >= 0 && case_x < bits_width &&
     case_y >= 0 && case_y < bits_height &&
     bits.(case_x).(case_y) <> 0 then
    begin
    (* Memorise la couleur actuelle *)
    current_color := bits.(case_x).(case_y) ;
    (* Rien n'est marque a l'origine *)
    marked_surface := 0 ;
    (* On fait le parcours de marquage *)
    mark_bits case_x case_y ;
    (* On met a jour le damier seulement si il y a plus d'1 composante *)
    (* connexe *)
    if !marked_surface > 1 then
     begin
     delete_bits () ;
     score := !score + (!marked_surface-2) * (!marked_surface-2) ;
     refresh_whole () ;
     refresh_score form;
     end
    else
     begin
     (* En fait, la surface est trop petite, donc remettre *)
     (* la case en etat *)
       sndPlaySystemSound 3;
     refresh_score form;
     bits.(case_x).(case_y) <- bits.(case_x).(case_y) land 0x7F
     end
    end
;;



(* ************************** *)
(* Declanchement d'une partie *)
(* ************************** *)
let newgame form =
 let screen_width = img_size * bits_width in
 let screen_height = img_size * bits_height + 30 in
 score := 0;
 init_bits () ;
 refresh_whole () ;
 refresh_score form
;;

let mainForm = 1000;;

let main_handler e =
  let form = frm__get_active_form() in
  match e with
  | FrmOpenEvent id ->
      frm__draw_form form;
      true
  | MenuEvent 101 ->
      newgame form;
      true
  | MenuEvent 102 -> freemem__f (); true
  | MenuEvent 103 -> freemem__gc(); true
  | PenDownEvent {x = x; y = y} ->
      busy form;
      handle_turn form (x / img_size) ((y - vshift) / img_size);
      if check_game_over() then	frm__alert 101;
      true
  | _ -> false
;;      

let main () =
  events__set_apphandler (function 
      FrmLoadEvent id -> 
	let form = frm__init_form id in
	if id = mainForm then begin
	  set_formhandler id main_handler
	end;
	frm__set_active_form form;
	true
    | FrmOpenEvent id ->
	frm__draw_form(frm__get_active_form());
	true
    | AppStopEvent ->
	print_debug "AppStop";
	false;
    | e -> (* do with other handlers *)
	print_event ser_debug e;
	false);
  frm__goto_form(mainForm)
;;

let _ = main ()
;;

  
