(* Buffered input and output *)

type in_channel;;
type out_channel;;
        (* The abstract types of input channels and output channels. *)

exception End_of_file
        (* Raised when an operation cannot complete, because the end
           of the file has been reached. *)
;;

(*** General output functions *)

value open_out : int -> out_channel = 1 "open_create"
        (* [open_descriptor_out fd] returns a buffered output channel
           writing to the file descriptor [fd]. The file descriptor [fd]
           must have been previously opened for writing, else the behavior is
	   undefined. *)
  and output_char : out_channel -> char -> unit = 2 "output_char"
        (* Write the character on the given output channel. *)
  and output_string : out_channel -> string -> unit
        (* Write the string on the given output channel. *)
  and output : out_channel -> string -> int -> int -> unit
        (* [output chan buff ofs len] writes [len] characters from string 
           [buff], starting at offset [ofs], to the output channel [chan].
           Raise [Invalid_argument "output"] if [ofs] and [len] do not
           designate a valid substring of [buff]. *)          
  and output_byte : out_channel -> int -> unit = 2 "output_char"
        (* Write one 8-bit integer (as the single character with that code)
           on the given output channel. The given integer is taken modulo
           256. *)
  and output_binary_int : out_channel -> int -> unit = 2 "output_int"
        (* Write one integer in binary format on the given output channel.
           The only reliable way to read it back is through the
           [input_binary_int] function. The format is compatible across
	   all machines for a given version of Caml Light. *)
  and output_value : out_channel -> 'a -> unit = 2 "extern_val"
        (* Write the representation of a structured value of any type
           to a channel. Circularities and sharing inside the value
           are detected and preserved. The object can be read back,
           by the function [input_value]. The format is compatible across
	   all machines for a given version of Caml Light. *)
  and seek_out : out_channel -> int -> unit = 2 "seek_out"
        (* [seek_out chan pos] sets the current writing position to [pos]
           for channel [chan]. This works only for regular files. On
           files of other kinds (such as terminals, pipes and sockets),
	   the behavior is unspecified. *)
  and pos_out : out_channel -> int = 1 "pos_out"
        (* Return the current writing position for the given channel. *)
  and out_channel_length : out_channel -> int = 1 "channel_size"
        (* Return the total length (number of characters) of the
           given channel.  This works only for regular files. On files of
           other kinds, the result is meaningless. *)
  and close_out : out_channel -> unit = 1 "close_out"
        (* Close the given channel, flushing all buffered write operations.
	   The behavior is unspecified if any of the functions above is
	   called on a closed channel. *)
;;

(*** General input functions *)

value open_in : int -> in_channel = 1 "open_atzero"
        (* [open_descriptor_in fd] returns a buffered input channel
           reading from the file descriptor [fd]. The file descriptor [fd]
           must have been previously opened for reading, else the behavior is
	   undefined. *)
  and input_char : in_channel -> char = 1 "input_char"
        (* Read one character from the given input channel.
           Raise [End_of_file] if there are no more characters to read. *)
  and input : in_channel -> string -> int -> int -> int
        (* [input chan buff ofs len] attempts to read [len] characters
           from channel [chan], storing them in string [buff], starting at
           character number [ofs]. It returns the actual number of characters
           read, between 0 and [len] (inclusive).
           A return value of 0 means that the end of file was reached.
           A return value between 0 and [len] exclusive means that
           no more characters were available at that time; [input] must be
           called again to read the remaining characters, if desired.
           Exception [Invalid_argument "input"] is raised if [ofs] and [len]
           do not designate a valid substring of [buff]. *)          
  and input_byte : in_channel -> int = 1 "input_char"
        (* Same as [input_char], but return the 8-bit integer representing
           the character.
           Raise [End_of_file] if an end of file was reached. *)
  and input_binary_int : in_channel -> int = 1 "input_int"
        (* Read an integer encoded in binary format from the given input
           channel. See [output_binary_int].
           Raise [End_of_file] if an end of file was reached while reading the
	   integer. *)
  and input_value : in_channel -> 'a = 1 "intern_val"
        (* Read the representation of a structured value, as produced
           by [output_value] or [output_compact_value], and return
           the corresponding value.
           This is not type-safe. The type of the returned object is
           not ['a] properly speaking: the returned object has one
           unique type, which cannot be determined at compile-time.
           The programmer should explicitly give the expected type of the
           returned value, using the following syntax:
                     [(input_value chan : type)].
	   The behavior is unspecified if the object in the file does not
	   belong to the given type. *)
  and seek_in : in_channel -> int -> unit = 2 "seek_in"
        (* [seek_in chan pos] sets the current reading position to [pos]
           for channel [chan]. This works only for regular files. On
           files of other kinds, the behavior is unspecified. *)
  and pos_in : in_channel -> int = 1 "pos_in"
        (* Return the current reading position for the given channel. *)
  and in_channel_length : in_channel -> int = 1 "channel_size"
        (* Return the total length (number of characters) of the
           given channel. This works only for regular files. On files of
           other kinds, the result is meaningless. *)
  and close_in : in_channel -> unit = 1 "close_in"
        (* Close the given channel. Anything can happen if any of the
           functions above is called on a closed channel. *)
;;
(*-*)
value open_or_create_in : int -> in_channel = 1 "open_create"
and open_or_create_out : int -> out_channel = 1 "open_create"
and   fast_input : in_channel -> string -> int -> int -> int = 4 "input"
and   fast_output : out_channel -> string -> int -> int -> unit = 4 "output"
and  palmio_init : string -> string -> string -> unit = 3 "clpalmio_init"
and  palmio_exit : unit -> unit = 1 "clpalmio_exit"
;;
