#include "mlvalues.h"
#include "fail.h"
#include "alloc.h"
#include "str.h"
#include "clPalm_cvt.h"
#include "sys.h"
/* 
 * Support for lists : the system absolutely needs a CharPtr * containing 
 * the list entries. We cannot give pointers to ML heap, because of garbage
 * collection. 
 * The proper solution is to make copies of the strings, and the array, and put
 * in ML a "finalized" object pointing to them. This object is kept in the 
 * ML heap, * (in a table associating ListPtr, o). When the choices are 
 * reset, the table is updated.
 * Currently, the finalized object is just an integer indicating a rsrc number.
 * Q: cleanup when the List object is destroyed ??
 */

static int make_list(value l, CharPtr **res, int *len)
{
  static Word counter = 0;
  VoidHand lh; 
  int datalen = 0; int ptrlen = 0;
  value car = l;
  /* First pass : compute length and length */
  while (Tag_val(car) == 1) {
    ptrlen++;
    datalen += string_length(Field(car, 0)) + 1; /* one for '\0' */
    car = Field(car, 1);
  };
  /* Allocate a resource : choose our own resource database, but reopen it
   * in read-write mode
   */
  {
    DmOpenRef db; LocalID id;
    DmSearchResource('CLby', 0, NULL, &db);
    if (db == NULL) sys_error(DmGetLastErr());
    DmOpenDatabaseInfo(db, &id, NULL, NULL, NULL, NULL);
    db = DmOpenDatabase(0, id, dmModeReadWrite);
    /* Allocate with proper size */
    lh = DmNewResource(db, 'CLll', counter++, datalen + sizeof(CharPtr) * ptrlen);
    if (lh == NULL) sys_error(DmGetLastErr());
    DmCloseDatabase(db);
  }
  /* Then we write the stuff */
  { 
    CharPtr ptr = MemHandleLock(lh); /* First lock the resource and get the pointer */
    ULong offset = sizeof(CharPtr) * ptrlen; /* the data are after the pointer array */
    int i = 0;
    car = l; 
    while(Tag_val(car) == 1) {
      char *pos = ptr + offset;	/* the position of the string copy */
      value s = Field(car, 0);	/* the string */
      DmWrite(ptr, offset, String_val(s), string_length(s) + 1); /* write the copy */
      DmWrite(ptr, i * sizeof(CharPtr), &pos, 4); /* write the pointer */
      car = Field(car, 1);	/* advance in the list */
      i++;			/* advance the pointer pointer */
      offset += string_length(s) + 1; /* advance the data */
    }
    *res = (CharPtr *)ptr;			/* pass beginning of array */
    *len = ptrlen;
  }
  return (counter - 1); /* index of resource */
}

static void free_list(int n)
{
  DmOpenRef db;
  LocalID id;
  Err err;
  Int index = DmSearchResource('CLll', n, NULL, &db);
  if (db == NULL) sys_error(DmGetLastErr());
  DmOpenDatabaseInfo(db, &id, NULL, NULL, NULL, NULL);
  db = DmOpenDatabase(0, id, dmModeReadWrite);
  err = DmRemoveResource(db, index);
  if (err != 0) sys_error(err);
  DmCloseDatabase(db);
}

value clpalm_Lstfreelist(value n) /* ML */
{
  free_list(Int_val(n));
  return Val_unit;
}


/*
 * Lst functions
 */

/* p308 : lst -> unit */
value clpalm_LstDrawList(value lptr) /* ML */
{
  LstDrawList((ListPtr)lptr);
  return Val_unit;
}

/* p308 : lst -> unit */
value clpalm_LstEraseList(value lptr) /* ML */
{
  LstEraseList((ListPtr)lptr);
  return Val_unit;
}

/* p309 : lst -> int  */
value clpalm_LstGetNumberOfItems(value lptr) /* ML */
{
  return Val_int(LstGetNumberOfItems((ListPtr)lptr));
}

/* p309 : lst -> int  */
value clpalm_LstGetVisibleItems(value lptr) /* ML */
{
  return Val_int(LstGetVisibleItems((ListPtr)lptr));
}


/* p310 : lst -> int  */
value clpalm_LstGetSelection(value lptr) /* ML */
{
  Word n = LstGetSelection((ListPtr)lptr);
  if (n == -1) mlraise(NOT_FOUND_EXN);
  return Val_int(n);
}

/* p310 : lst -> int -> string  */
value clpalm_LstGetSelectionText(value lptr, value idx) /* ML */
{
  CharPtr p = LstGetSelectionText((ListPtr)lptr, Int_val(idx)); 
  if (p == NULL) mlraise(NOT_FOUND_EXN);
  return copy_string(p);
}

value clpalm_LstHandleEvent(value lptr, value ev) /* ML */
{
  EventType e;
  clpalm_mlevent2c(ev, &e);
  return Val_bool(LstHandleEvent((ListPtr)lptr, &e));
}

/* p312 : lst -> int -> unit */
value clpalm_LstMakeItemVisible(value lptr, value idx) /* ML */
{
  LstMakeItemVisible((ListPtr)lptr, Int_val(idx));
  return Val_unit;
}

/* p312 : lst -> int */
value clpalm_LstPopupList(value lptr) /* ML */
{
  short n = LstPopupList((ListPtr)lptr);
  if (n == -1) mlraise(NOT_FOUND_EXN);
  return Val_int(n);
}

/* p313 : lst -> direction -> int -> bool */
value clpalm_LstScrollList(value lptr, value direction, value itemcount) /* ML */
{
  return Val_bool(LstScrollList((ListPtr)lptr, Tag_val(direction),
				Int_val(itemcount)));
}

/* TODO p323: LstSetDrawFunction - understand the dispatch */

/* p314: lst -> int -> unit */
value clpalm_LstSetHeight(value lptr, value visible) /* ML */
{
  LstSetHeight((ListPtr)lptr, Int_val(visible));
  return Val_unit;
}

/* p314 : lst -> string array -> unit */
/* We do have a problem here: if we put the array pointer in the List object, it
 * might become invalid (unless we tell the GC about it being a root).
 * And then, if the List object is actually in protected memory (as is probably the
 * case with resources), the GC won't be able to update the pointer.
 */
value clpalm_LstSetListChoices(value lptr, value choices) /* ML */
{
  CharPtr *ptr; int len;
  int n = make_list(choices, &ptr, &len);
  LstSetListChoices((ListPtr)lptr, ptr, len);
  return Val_int(n);
}

/* p315: lst -> int -> int -> unit */
value clpalm_LstSetPosition(value lptr, value x, value y) /* ML */
{
  LstSetPosition((ListPtr)lptr, Int_val(x), Int_val(y));
  return Val_unit;
}

/* p315 :  lst -> int -> unit */
value clpalm_LstSetSelection(value lptr, value item) /* ML */
{
  LstSetSelection((ListPtr)lptr, Int_val(item));
  return Val_unit;
}

/* p315 : lst  -> int -> unit */
value clpalm_LstSetTopItem(value lptr, value item) /* ML  */
{
  LstSetTopItem((ListPtr)lptr, Int_val(item));
  return Val_unit;
}

