#include "alloc.h"
#include "fail.h"
#include "memory.h"
#include "mlvalues.h"

/* PalmOS specific float support:
 * REQUIRES: MathLib.prc, from Rick Huebner
 * We go through Flp routines only for the primitive
 * operations +,-,*,/ and string conversion. The rest
 * uses standard double.
 */

/* See below for this */
#define _DONT_USE_FP_TRAPS_ 1

#include "MathLib.h"

int clpalmfp_init()
{
  Err error;

  error = SysLibFind(MathLibName, &MathLibRef);
  if (error)
    error = SysLibLoad(LibType, MathLibCreator, &MathLibRef);
  if (error) {
    ErrDisplay("Can't find MathLib");
    return 1;
  }
  error = MathLibOpen(MathLibRef, MathLibVersion);
  if (error) {
    ErrDisplay("Can't open MathLib"); 
    return 1;
  }
  return 0;
}

void clpalmfp_close()
{
  Err error;
  UInt usecount;

  error = MathLibClose(MathLibRef, &usecount);
  if (error)
    ErrDisplay(error, "Can't close MathLib");
  if (usecount == 0)
SysLibRemove(MathLibRef); 
}

/* Since floats are emulated, we assume they can be word aligned,
 * and not necessarily double-word aligned. If this is not correct,
 * change .../config/m.h.
 * In either case, we keep the following code in this file.
 */

#ifdef CAML_ALIGN_DOUBLE
double Double_val(val)
     value val;
{
  union { value v[2]; double d; } buffer;

  Assert(sizeof(double) == 2 * sizeof(value));
  buffer.v[0] = Field(val, 0);
  buffer.v[1] = Field(val, 1);
  return buffer.d;
}

void Store_double_val(val, dbl)
     value val;
     double dbl;
{
  union { value v[2]; double d; } buffer;

  Assert(sizeof(double) == 2 * sizeof(value));
  buffer.d = dbl;
  Field(val, 0) = buffer.v[0];
  Field(val, 1) = buffer.v[1];
}

#endif

/* We don't support format_float yet */

/* The following hacks seem necessary with prc-tools 0.5.0
 * because of different calling conventions with structures
 * Hacks proposed by Jack Rouse <bratwrst@netcom.com>
 */

void SysTrapFlpAToF(FlpDouble*, Char*) SYS_TRAP(sysTrapFlpDispatch);
// note there are two different NewFloatMgr related system traps

double SysFlpAToF( char* s )
{
    FlpCompDouble fcd;
    asm("moveq.l %0,%%d2" : : "i" (sysFloatAToF) : "d2");
    SysTrapFlpAToF(&fcd.fd, s);
    return fcd.d;
}
/* NOTE: this is probably not working due to
 * different conventions for returning structures
 */

value float_of_string(s)        /* ML */
     value s;
{
  FlpCompDouble x;
  x.fd = SysFplAToF(String_val(s));
  return copy_double(x.d);
}
/* SDK sez:
   FplFToA 
   Convert a floating-point number to a zero-terminated
   ASCII string in exponential format : [-]x.yyyyyyyye[-]zz
   Thus the size of the result is at most 16;
*/
value string_of_float(value f)	/* ML */
{
  char buf[16];
  FlpCompDouble x; /* union of double/FlpDouble */
  int err;
  x.d = Double_val(f);
  err = FlpFToA(x.fd, buf);
  if (err)
    invalid_argument("FlpFToA");
  return copy_string(buf);
}

value exp_float(f)              /* ML */
     value f;
{
  return copy_double(exp(Double_val(f)));
}

value fabs_float(f)              /* ML */
     value f;
{
  return copy_double(fabs(Double_val(f)));
}

value floor_float(f)              /* ML */
     value f;
{
  return copy_double(floor(Double_val(f)));
}

value fmod_float(f1, f2)              /* ML */
     value f1, f2;
{
  return copy_double(fmod(Double_val(f1), Double_val(f2)));
}

value frexp_float(f)              /* ML */
     value f;
{
  int i;
  value res;
  Push_roots(r, 1);

  r[0] = copy_double(frexp (Double_val(f), &i));
  res = alloc_tuple(2);
  Field(res, 0) = r[0];
  Field(res, 1) = Val_int(i);
  Pop_roots();
  return res;
}

value ldexp_float(f, i)              /* ML */
     value f, i;
{
  return copy_double(ldexp(Double_val(f), Int_val(i)));
}

value log_float(f)              /* ML */
     value f;
{
  return copy_double(log(Double_val(f)));
}

value log10_float(f)              /* ML */
     value f;
{
  return copy_double(log10(Double_val(f)));
}

value modf_float(f)              /* ML */
     value f;
{
  double fres;
  value res;
  Push_roots(r, 2);

  r[0] = copy_double(modf (Double_val(f), &fres));
  r[1] = copy_double(fres);

  res = alloc_tuple(2);
  Field(res, 0) = r[0];
  Field(res, 1) = r[1];
  Pop_roots();
  return res;
}

value sqrt_float(f)             /* ML */
     value f;
{
  return copy_double(sqrt(Double_val(f)));
}

value power_float(f, g)         /* ML */
     value f, g;
{
  return copy_double(pow(Double_val(f), Double_val(g)));
}

value sin_float(f)              /* ML */
     value f;
{
  return copy_double(sin(Double_val(f)));
}

value sinh_float(f)              /* ML */
     value f;
{
  return copy_double(sinh(Double_val(f)));
}

value cos_float(f)              /* ML */
     value f;
{
  return copy_double(cos(Double_val(f)));
}

value cosh_float(f)              /* ML */
     value f;
{
  return copy_double(cosh(Double_val(f)));
}

value tan_float(f)              /* ML */
     value f;
{
  return copy_double(tan(Double_val(f)));
}

value tanh_float(f)              /* ML */
     value f;
{
  return copy_double(tanh(Double_val(f)));
}

value asin_float(f)             /* ML */
     value f;
{
  return copy_double(asin(Double_val(f)));
}

value acos_float(f)             /* ML */
     value f;
{
  return copy_double(acos(Double_val(f)));
}

value atan_float(f)             /* ML */
     value f;
{
  return copy_double(atan(Double_val(f)));
}

value atan2_float(f, g)        /* ML */
     value f, g;
{
  return copy_double(atan2(Double_val(f), Double_val(g)));
}

value ceil_float(f)              /* ML */
     value f;
{
  return copy_double(ceil(Double_val(f)));
}
