#include "config.h"
#include "mlvalues.h"
#include "memory.h"
#include "fail.h"
#include "sys.h"
#include "alloc.h"
#include "io.h"

unsigned getblock (struct channel *chan, char *dst, unsigned length)
{
  long available = chan->max - chan->offset;
  unsigned long red;
  char *src = (char *)MemHandleLock(chan->record) + chan->offset;

  // printf("getblock %lx\n", src);
  if (length <= available) 
    red = length;
  else if (available > 0)
    red = available;
  else 
    red = 0;
  if (red) {
    bcopy(src, dst, red);
    chan->offset += red;
  }
  
  MemHandleUnlock(chan->record);
  return red;
}

uint32 getword(chan)
     struct channel * chan;
{
  long available = chan->max - chan->offset;
  unsigned char *src = 
    (unsigned char *)MemHandleLock(chan->record) + chan->offset;
  
  if (available < 4) {
    MemHandleUnlock(chan->record);
    mlraise(Atom(END_OF_FILE_EXN));
  } else {
    uint32 res = 0;
    int i;

    for(i = 0; i < 4; i++, chan->offset++) {
      res = (res << 8) + src[i];
    }
    MemHandleUnlock(chan->record);
    return res;
  }
}

/* Since we don't buffer, this is actually quite costly 
 * (lock/unlock for each character)
 */
value input_char (struct channel *chan)	/* ML */
{
  if (chan->offset < chan->max) {
    int c;
    char *src = (unsigned char *)MemHandleLock(chan->record) + chan->offset;
    chan->offset++;
    c = *src;
    MemHandleUnlock(chan->record);
    return Val_int(c);
  } else
    mlraise(Atom(END_OF_FILE_EXN));
}


value input (value channel, value buff, value start, value length) /* ML */
{
  return Val_long(getblock((struct channel *) channel,
                           &Byte(buff, Long_val(start)),
                           (ULong) Long_val(length)));
}

value seek_in(struct channel *chan, value pos)	/* ML */
{
  long newpos = Long_val(pos);

  if ((newpos < 0) || (newpos > chan->max))
    sys_error(dmErrSeekFailed);	/* TODO check the semantics */
  else {
    chan->offset = newpos;
    return Val_unit;
  }
}

/* Until we understand exactly how record numbering works, we don't trust
 * chan->index. Instead, we search again in the DB.
 * Could use uniqueID as well.
 */
value close_in(struct channel *chan)	/* ML */
{
  if (chan->resource == false) {
    DmOpenRef adm;
    chan->index = DmSearchRecord(chan->record, &adm);
    if (chan->index == -1) sys_error(DmGetLastErr());
    DmReleaseRecord(adm,chan->index,false);
  }
  else
    DmReleaseResource(chan->record);
  stat_free((char *)chan);
  return Val_unit;   
}

value channel_size(struct channel *chan) /* ML */
{
  return Val_long(chan->max);
}

value channel_pos(struct channel *chan)	/* ML */
{
  return Val_long(chan->offset);
}

static void resize_chan(struct channel *chan, ULong size)
{
  VoidHand newhandle;
  DmOpenRef adm;

  chan->index = DmSearchRecord(chan->record, &adm);
  if (chan->index == -1) sys_error(DmGetLastErr());
  newhandle = DmResizeRecord(adm, chan->index, size);
  if (newhandle == NULL) sys_error(DmGetLastErr());
  chan->record = newhandle;
}

void putblock(struct channel *chan, char *src, unsigned bytes)
{
  VoidPtr recordP;
  ULong current_size = MemHandleSize(chan->record);

  if (bytes + chan->offset < current_size) {
    recordP = MemHandleLock(chan->record);
    DmWrite(recordP, chan->offset, src, bytes);
    chan->offset += bytes;
    chan->max = max(chan->offset, chan->max);
    MemHandleUnlock(chan->record);
  } else { /* out of bounds */
    resize_chan(chan, current_size + DEFAULT_SIZE);
    putblock(chan, src, bytes);	/* try again, recursively */
  }
}

void putword(struct channel *chan, uint32 w)
{
  unsigned char buf[4];
  buf[0] = (w >> 24) & 0xff;
  buf[1] = (w >> 16) & 0xff;
  buf[2] = (w >> 8) & 0xff;
  buf[3] = w & 0xff;
  putblock(chan, buf, 4);
}


/* overkill, but I don't want to copy this code */
value output_char(struct channel *chan, value c) /* ML */
{
  unsigned char buf[1];
  buf[0] = Long_val(c);
  putblock(chan, buf, 1);
  return Atom(0);
}


value output(channel, buff, start, length) /* ML */
     value channel, buff, start, length;
{
  putblock((struct channel *) channel,
           &Byte(buff, Long_val(start)),
           (ULong) Long_val(length));
  return Atom(0);
}

value seek_out(struct channel *chan, value pos)	/* ML */
{
  long newpos = Long_val(pos);
  ULong current_size = MemHandleSize(chan->record);

  if (newpos < 0)
    sys_error(dmErrSeekFailed);	/* TODO check the semantics */
  if (newpos >= current_size)
    resize_chan(chan, newpos + DEFAULT_SIZE);

  chan->offset = newpos;
  chan->max = max(chan->max, chan->offset);
  return Val_unit;
}

/* see beginning of file for explanation of Resize */
value close_out(struct channel *chan) /* ML */
{
  if (chan->resource == true) {
    /* not supported yet */
    DmReleaseResource(chan->record);
  } else {
    DmOpenRef adm;
    resize_chan(chan, chan->max);
    chan->index = DmSearchRecord(chan->record, &adm);
    if (chan->index == -1) sys_error(DmGetLastErr());
    DmReleaseRecord(adm,chan->index,false);
  }
  stat_free((char *)chan);
  return Val_unit;   
}

/* open a channel on a resource handle 
 * Currently, we assume resources are binary files because bytecode and global data
 * reside in a resource DB.
 * TODO: support text mode
 */
struct channel *open_descr(VoidHand h)
{
  struct channel *chan = 
    (struct channel *) stat_alloc (sizeof(struct channel));
  chan->record = h;
  chan->index = 0;
  chan->offset = 0L;
  chan->resource = true;
  chan->binary = true;
  chan->max = MemHandleSize(h);	/* since we assume binary mode */
  return chan;
}

