/*
    Jump - Java post-compiler for PalmPilot
    Copyright (C) 1996,97  Greg Hewgill <gregh@lightspeed.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
    This file has been modified in 03-05/2000 
    by Ralf Kleberhoff <Kleberhoff@aol.com>
    to make Jump compatible with current Java and PalmOS versions
    and to extend the functionality.
    Many thanks to Greg Hewgill for the original version!
*/

import java.io.*;
import java.util.*;

class AsmBackEnd implements BackEnd, JVM {

  /** the output stream for the asm file. */
  PrintWriter asmout;

  /** the filename of the asm file. */
  String asmoutFilename;

  /** the project name */
  String projectname;

  /** counter for producing unique hash keys. */
  static int depCounter = 0;

  /** Hashtable of natives found on the path. */
  Hashtable allNatives;

  /** The length of Itables in words */
  int itableWords = 0;

  /*
   *          Constructors and related methods
   * ====================================================
   */

  /**
   * construct an AsmBackEnd from project name and directory.
   */
  AsmBackEnd (String projectname, String baseDirectory)
  {
    File out = new File (baseDirectory, projectname + ".asm");

    asmoutFilename = out.getPath();
    this.projectname = projectname;
  }

  /*
   *             natives inclusion methods
   * ====================================================
   */

  /**
   * insert native methods into a class.
   * If a java version (from the class file) exists, it is replaced
   * by the native method.
   */
  public void insertNatives (Klass cls)
  {
    loadAllNatives();

    if (cls.methods != null) {
      for (int i=0; i<cls.methods.length; i++) {
	MethodInfo method = cls.methods[i];
	NativeRef ref = (NativeRef) allNatives.get (method.fullName());

	// allow for wildcard as in '*.registerNatives()V'
	if (ref == null) {
	  String wildname = "*." + method.name + method.signature;
	  ref = (NativeRef) allNatives.get (wildname);
	}

	if (ref != null) {
	  Vector deps = new Vector();
	  for (Enumeration enum=ref.dependencies.elements();
	       enum.hasMoreElements(); ) {
	    String element = (String) enum.nextElement();
	    int dotPos = element.indexOf (".");
	    int parenPos = element.indexOf ("(");

	    if (element.startsWith ("new ")) {
	      deps.addElement 
		(new DependencyInfo (Klass.forName (element.substring (4)),
				     JavaElement.INSTANCE_NEEDED));
	    }
	    if (element.startsWith ("exact ")) {
	      deps.addElement 
		(new DependencyInfo (Klass.forName (element.substring (6)),
				     JavaElement.EXACT_INSTANCE_NEEDED));
	    }
	    else if (dotPos < 0) {
	      deps.addElement (Klass.forName (element));
	    }
	    else if (parenPos < 0) {
	      String cln = element.substring (0, dotPos);
	      String fln = element.substring (dotPos+1);

	      deps.addElement (Klass.forName(cln).findField(fln));
	    }
	    else {
	      String cln = element.substring (0, dotPos);
	      String mtn = element.substring (dotPos+1, parenPos);
	      String mts = element.substring (parenPos);

	      deps.addElement (Klass.forName(cln).findMethod(mtn,mts));
	    }
	  }
	  method.makeNative (ref, deps);
	}
      }
    }
  }

  /**
   * update the 'needed' state of all elements
   * that the back-end's implementation depends upon.
   */
  public void updateAll()
  {
    Klass stringClass = Klass.forName ("java/lang/String");
    FieldInfo vField = stringClass.findField ("value");
    FieldInfo cField = stringClass.findField ("count");
    FieldInfo oField = stringClass.findField ("offset");
	    
    stringClass.markNeeded ("AsmBackEnd", JavaElement.INSTANCE_NEEDED);
    if (vField != null) vField.markNeeded ("AsmBackEnd", JavaElement.NEEDED);
    if (cField != null) cField.markNeeded ("AsmBackEnd", JavaElement.NEEDED);
    if (oField != null) oField.markNeeded ("AsmBackEnd", JavaElement.NEEDED);

    Klass.forName ("[C").
      markNeeded ("AsmBackEnd", JavaElement.INSTANCE_NEEDED);

    try {
      ((NativeRef) allNatives.get ("kernel-routines")).
	markNeeded("kernel", JavaElement.NEEDED);
    }
    catch (NullPointerException e) { }

    boolean[] used = Klass.getAllBytecodesUsed();
    for (int i=0; i<used.length; i++) {
      if (used[i]) {
	try {
	  ((NativeRef) allNatives.get ("op_" + OPCODE_NAMES[i])).
	    markNeeded("opcode", JavaElement.NEEDED);
	}
	catch (NullPointerException e) { }
      }
    }

    for (Enumeration enum=allNatives.elements();
	 enum.hasMoreElements(); ) {
      ((NativeRef) enum.nextElement()).updateNeeded();
    }
  }

  /**
   * load all native methods.
   */
  void loadAllNatives ()
  {
    if (allNatives == null) {
      allNatives = new Hashtable();
      
      String nativelist = (Jump.projectProps.getProperty ("jump-natives", "") +
			   " " + Jump.projectProps.getProperty ("natives", ""));
      
      System.out.println ("Loading natives and opcodes from " + nativelist);

      StringTokenizer stn = new StringTokenizer (nativelist);
      Vector natives = new Vector();
      while (stn.hasMoreTokens()) {
	natives.addElement (stn.nextToken());
      }

      for (int i=0; i<natives.size(); i++) {
	String file = (String) natives.elementAt(i);

	if (Jump.Verbosity >= 1) {
	  System.out.println("Reading " + file);
	}
	try {
	  LineNumberReader nf = (new LineNumberReader 
				 (new InputStreamReader 
				  (ClassLoader.getSystemResourceAsStream (file))));
	  NativeRef currentRef = null;
	  while (true) {
	    String s = nf.readLine();
	    StringTokenizer st;
	    Vector tokens;
	    int pos;

	    if (s == null) {
	      break;
	    }
	    if (s.startsWith (";;")) {
	      try {
		st = new StringTokenizer(s);
		tokens = new Vector();
		while (st.hasMoreTokens()) {
		  tokens.addElement(st.nextToken());
		}

		if ((tokens.size() == 2) &&
		    (tokens.elementAt(1).equals("kernel-routines"))) {
		  currentRef = new NativeRef ("kernel-routines", file);
		  allNatives.put ("kernel-routines", currentRef);
		}
		else if ((tokens.size() == 3) &&
			 (tokens.elementAt(1).equals("native-method"))) {
		  String sig = (String) tokens.elementAt(2);
		  currentRef = new NativeRef (sig, file, nf.getLineNumber());
		  allNatives.put (sig, currentRef);
		}
		else if ((tokens.size() == 3) &&
			 (tokens.elementAt(1).equals("jvm-opcode"))) {
		  String opcodeName = "op_" + (String) tokens.elementAt(2);
		  currentRef = new NativeRef (opcodeName, file, nf.getLineNumber());
		  allNatives.put (opcodeName, currentRef);
		}
		else if ((tokens.size() == 5) &&
			 (tokens.elementAt(1).equals("uses-class")) &&
			 (tokens.elementAt(3).equals("as"))) {
		  String sig = (String) tokens.elementAt(2);
		  String sym = (String) tokens.elementAt(4);
		  currentRef.addDependency (sig, sym);
		}
		else if ((tokens.size() == 3) &&
			 (tokens.elementAt(1).equals("uses-class"))) {
		  String sig = (String) tokens.elementAt(2);
		  currentRef.addDependency (sig, null);
		}
		else if ((tokens.size() == 3) &&
			 (tokens.elementAt(1).equals("uses-instance"))) {
		  String sig = (String) tokens.elementAt(2);
		  currentRef.addDependency ("new " + sig, null);
		}
		else if ((tokens.size() == 5) &&
			 (tokens.elementAt(1).equals("uses-method")) &&
			 (tokens.elementAt(3).equals("as"))) {
		  String sig = (String) tokens.elementAt(2);
		  String sym = (String) tokens.elementAt(4);
		  currentRef.addDependency (sig, sym);
		}
		else if ((tokens.size() == 3) &&
			 (tokens.elementAt(1).equals("uses-method"))) {
		  String sig = (String) tokens.elementAt(2);
		  currentRef.addDependency (sig, null);
		}
		else if ((tokens.size() == 5) &&
			 (tokens.elementAt(1).equals("uses-field")) &&
			 (tokens.elementAt(3).equals("as"))) {
		  String sig = (String) tokens.elementAt(2);
		  String sym = (String) tokens.elementAt(4);
		  currentRef.addDependency (sig, sym);
		}
		else if ((tokens.size() == 3) &&
			 (tokens.elementAt(1).equals("uses-field"))) {
		  String sig = (String) tokens.elementAt(2);
		  currentRef.addDependency (sig, null);
		}
		else if ((tokens.size() == 3) &&
			 (tokens.elementAt(1).equals("needs-exact-layout"))) {
		  String sig = (String) tokens.elementAt(2);
		  currentRef.addDependency ("exact " + sig, null);
		}
	      }
	      catch (Exception e) {
		System.out.println ("syntax error in file " + file + ": " + s);
		e.printStackTrace();
	      }
	    }
	  }
	  nf.close();
	} 
	catch (Exception e) {
	  System.out.println ("error in file " + file);
	  e.printStackTrace();
	}
      }
    }
  }

  /**
   * copy the contents of an open stream to the assembly output stream.
   */
  void copyStream (InputStream stream, String name)
  {
    try {
      BufferedReader in = new BufferedReader (new InputStreamReader (stream));
      String line;
      
      asmout.println ();
      asmout.println (";; --- from file " + name + " ---");
      asmout.println ();
      
      while ((line = in.readLine()) != null) {
	asmout.println (line);
      }
      in.close();
    }
    catch (Exception e) {
      e.printStackTrace();
    }
    asmout.println ();
  }

  /*
   *             code generation methods
   * ====================================================
   */

  /**
   * generate the code.
   */
  public void generate()
  {
    try {
      assertCompatibility();

      asmout = new PrintWriter (new BufferedWriter
				(new FileWriter (asmoutFilename)));
      
      generateHeader();
      generateClassTable();
      generateVtableAll();
      generateStatic();
      generatePilotMain();
      generateClinitCode();

      for (int i=0; i<Klass.ClassList.size(); i++) {
	Klass cl = (Klass) Klass.ClassList.elementAt(i);
	if (cl.methods != null) {
	  for (int j=0; j<cl.methods.length; j++) {
	    MethodInfo m = cl.methods[j];
	    if (m.isNeeded(JavaElement.NEEDED)) {
	      generateMethod (m);
	    }
	  }
	}
      }
      
      generateConstantStrings();
      generateSubroutines();

      generateResources();

      generateLeadout();
      
      asmout.close();
    }
    catch (Exception e) {
      e.printStackTrace();
    }
  }

  /**
   * write the (more or less constant) header part of the asm file.
   */
  void generateHeader ()
  {
    String appName = 
      new String (Jump.projectProps.getProperty ("appname", projectname));
    String appId = 
      new String (Jump.projectProps.getProperty ("appid", "test"));

    asmout.println("; Generated from " + projectname + " by Jump");
    asmout.println();
    asmout.println("\tappl\t\"" + appName + "\",'" + appId + "'");
    asmout.println("\tinclude\t\"Pilot.inc\"");
    asmout.println("\tinclude\t\"startup.inc\"");
    asmout.println();

    copyStream (ClassLoader.getSystemResourceAsStream ("define.asm"), 
		"define.asm");
  }

  /**
   * write one Itable.
   */
  void generateItable (boolean[] itable)
  {
    int itableBits = 16 * itableWords;

    asmout.println ("\t;; Itable");
    for (int i=0; i<itableBits; i+=8) {
      int byteval = 0;
      for (int j=0; j<8; j++) {
	if ((i+j < itable.length) && itable[i+j]) {
	  byteval += (1 << j);
	}
      }
      asmout.println ("\tdc.b\t%" + Integer.toBinaryString (byteval));
    }
  }


  /**
   * write one class table entry to the asm file.
   */
  void generateClassTableEntry (Klass cl)
  {
    // a comment naming the class
    asmout.println ("\t; " + cl);
    if (cl.superclass != null) {
      asmout.println("\t; extends " + cl.superclass.className);
    }

    // the ObjectHeader_ClassIndex for the fake Class object
    asmout.println("\tdc.w\t1\t\t; ObjectHeader_ClassIndex: java/lang/Class");

    // the class-index of the superclass
    if (cl.superclass != null) {
      asmout.println("\tdc.w\t" + cl.superclass.classIndex + 
		     "\t; " + cl.superclass.className);
    } else {
      asmout.println("\tdc.w\t-1\t; (no superclass)");
    }

    // the size in bytes of an instance of this class
    asmout.println("\tdc.w\tObjectHeader_sizeof+" + cl.dataSize);

    // the class type (0=instance, 1=scalar array, 2=object array
    if ((cl.className.startsWith ("[L")) || (cl.className.startsWith ("[[")))  {
      asmout.println("\tdc.w\t2");
    }
    else if (cl.className.startsWith ("["))  {
      asmout.println("\tdc.w\t1");
    }
    else {
      asmout.println("\tdc.w\t0");
    }

    // the name of this class
    asmout.println("\tdc.l\tclass" + cl.classIndex + "__name");

    // the vtable and vtable size for this class
    if (cl.vtable != null) {
      asmout.println("\tdc.l\tclass" + cl.classIndex + "__vtable");
    }
    else {
      asmout.println("\tdc.l\t0\t\t; no vtable");
    }

    // the ArrayStoreException test code
    // (only for array classes of reference types, having instances)
    if ((cl.className.startsWith ("[[") ||
	 cl.className.startsWith ("[L")) &&
	cl.isNeeded (JavaElement.INSTANCE_NEEDED)) {
      Klass elClass = Klass.forSignature (cl.className.substring(1));
      int bitnum = elClass.instanceofClassIndex;
      if (bitnum < 0) {
	ASSERT.fail ("'aastore' check code for " + cl + " not prepared!");
      }

      asmout.println("\tbtst\t#" +  bitnum%8 + 
		     ",ClassInfo.Itable+" + bitnum/8 + "(a3)");
      asmout.println("\trts");
    }
    else {
      asmout.println("\tdc.w\t0,0,0,0\t\t; no itable index");
    }

    if (cl.itable != null) {
      generateItable (cl.itable);
    }
    else {
      for (int i=0; i<itableWords; i++) {
	asmout.println ("\tdc.w\t0" + (i==0 ? "\t\t; no itable" : ""));
      }
    }
    asmout.println();
  }

  /**
   * write the class table to the asm file.
   */
  void generateClassTable ()
  {
    if (Klass.instanceofTargetClassList != null) {
      itableWords = (Klass.instanceofTargetClassList.size() + 15) / 16;
    }

    asmout.println("\tcode");
    asmout.println();
    asmout.println("ClassInfo_size\tequ\tClassInfo_basesize+" + 
		   (2 * itableWords));
    asmout.println();
    asmout.println("\talign\t2");
    asmout.println("ClassTable:");
    for (int i=0; i<Klass.targetClassList.size(); i++) {
      Klass cl = (Klass) Klass.targetClassList.elementAt(i);
      generateClassTableEntry (cl);
    }
  }

  /**
   * write the vtable and the classname for this class
   */
  void generateVtable (Klass cl)
  {
    asmout.println ("class" + cl.classIndex + 
		    "__name\tdc.b\t'" + cl.className + "',0");
    asmout.println ("\talign\t2");

    if (cl.vtable != null) {
      asmout.println ("class" + cl.classIndex + "__vtable:");
      for (int i=0; i<cl.vtable.size(); i++) {
	MethodInfo m = (MethodInfo) cl.vtable.elementAt(i);
	if (m != null) {
	  asmout.println("\tdc.l\t" + m.shortLabel() + "\t; " + m.toString());
	}
	else {
	  asmout.println("\tdc.l\tM_none\t; (unused vtable entry)");
	}
      }
    }
    
    asmout.println();
  }

  /**
   * write classname and (if needed) vtable for all classes.
   */
  void generateVtableAll ()
  {
    for (int i=0; i<Klass.targetClassList.size(); i++) {
      Klass cl = (Klass) Klass.targetClassList.elementAt(i);
      generateVtable (cl);
    }
  }

  /**
   * write the section for the static fields.
   */
  void generateStatic ()
  {
    asmout.println("StaticObjectCount\tdc.w\t" + Klass.neededObjectFields.size());
    asmout.println();
    asmout.println("\tdata");
    asmout.println();
    asmout.println("StackEnd\tds.l\t1");
    asmout.println("FirstObject\tds.l\t1");
    asmout.println();
    asmout.println("StaticObjects:");
    asmout.println();
    
    // object fields
    asmout.println ("\t; static object fields");
    for (int i=0; i<Klass.neededObjectFields.size(); i++) {
      FieldInfo field = (FieldInfo) Klass.neededObjectFields.elementAt(i);

      asmout.println (field.shortLabel() + "\tds.l\t1\t; " + field);
    }
    asmout.println();

    // scalar one-byte fields
    asmout.println ("\t; static scalar one-byte fields");
    for (int i=0; i<Klass.neededScalarFields.size(); i++) {
      FieldInfo field = (FieldInfo) Klass.neededScalarFields.elementAt(i);

      switch (field.signature.charAt(0)) {
      case 'Z':
      case 'B':
      case 'C':
	asmout.println (field.shortLabel() + "\tds.b\t1\t; " + field);
	break;
      }
    }
    asmout.println();

    // scalar multi-byte fields
    asmout.println ("\t; static scalar multi-byte fields");
    asmout.println ("\talign\t2");
    for (int i=0; i<Klass.neededScalarFields.size(); i++) {
      FieldInfo field = (FieldInfo) Klass.neededScalarFields.elementAt(i);

      switch (field.signature.charAt(0)) {
      case 'S':
	asmout.println (field.shortLabel() + "\tds.w\t1\t; " + field);
	break;
      case 'F':
      case 'I':
	asmout.println (field.shortLabel() + "\tds.l\t1\t; " + field);
	break;
      case 'D':
      case 'J':
	asmout.println (field.shortLabel() + "\tds.l\t2\t; " + field);
	break;
      }
    }
  }

  /**
   * write the constant strings. <p>
   *
   * This is:
   * <ul>
   * <li>One big character block in the code segment,
   *     containing all the constant strings,
   *     always terminated with a Null character.</li>
   * <li>A fake character-array object in the data segment,
   *     pointing to this block.</li>
   * <li>Many fake String objects in the data segment,
   *     pointing (with offset) into the character array.</li>
   * <li>An 'init_constant_strings' routine to set up the pointers.</li>
   * </ul>
   */
  void generateConstantStrings ()
  {
    String[] strings = ConstantStrings.getStringArray();
    int totalLength = 0;

    // write the character block
    asmout.println ("cstring_block:");
    for (int i=0; i<strings.length; i++) {
      String s = strings[i];
      int len = s.length();
      if (len > 0) {
	asmout.print ("\t;; \"" + commentFilter(s) + "\"");
	for (int j=0; j<len; j++) {
	  if ((j & 15) == 0) {
	    asmout.println ();
	    asmout.print ("\tdc.b\t$" + Integer.toHexString ((int) s.charAt (j)));
	  }
	  else {
	    asmout.print (",$" + Integer.toHexString ((int) s.charAt (j)));
	  }
	}
	asmout.println (", 0");
      }
      else {
	asmout.println ("\t;; empty string");
	asmout.println ("\tdc.b\t0");
      }
      asmout.println();
      totalLength += len+1;
    }
    asmout.println ("\talign\t2");
    asmout.println();

    // write the fake character-array object
    asmout.println ("\tdata");
    asmout.println ("\tdc.w\t" + Klass.forName("[C").classIndex);
    asmout.println ("cstring_array:");
    asmout.println ("\tdc.w\t" + totalLength + "\t; length");
    asmout.println ("\tdc.w\t1\t; elsize");
    asmout.println ("\tdc.w\t0\t; objtype");
    asmout.println ("cstring_array_data:");
    asmout.println ("\tdc.l\t0\t; data");
    asmout.println();
    
    // write the fake String objects
    int index = 0;
    Klass stringClass = Klass.forName ("java/lang/String");
    int stringClassIndex = stringClass.classIndex;
    FieldInfo[] fields = stringClass.fields;
    FieldInfo vField = stringClass.findField ("value");
    FieldInfo cField = stringClass.findField ("count");
    FieldInfo oField = stringClass.findField ("offset");
    
    for (int i=0; i<strings.length; i++) {
      String s = strings[i];
      int len = s.length();
      String label = ConstantStrings.getLabel (s);
      
      asmout.println ("\t;; \"" + commentFilter(s) + "\"");
      asmout.println ("\tdc.w\t" + stringClassIndex);
      asmout.println (label + ":");

      boolean odd = false;
      for (int j=0; j<fields.length; j++) {
	FieldInfo f = fields[j];
	if (f.isNeeded() && ((f.access_flags & ACC_STATIC) == 0)) {
	  if (odd && f.size > 1) {
	    asmout.println ("\talign\t2");
	    odd = false;
	  }
	  if (f == vField) {
	    asmout.println (label + "_value:");
	    asmout.println ("\tdc.l\t0\t; 'value' points to cstring_array");
	  }
	  else if (f == oField) {
	    asmout.println ("\tdc.l\t" + index + "\t; 'offset' within cstring_array");
	  }
	  else if (f == cField) {
	    asmout.println ("\tdc.l\t" + len + "\t; 'count' String length");
	  }
	  else if (f.size == 1) {
	    asmout.println ("\tdc.b\t0\t; '" + f.name + "'");
	    odd = !odd;
	  }
	  else if (f.size == 2) {
	    asmout.println ("\tdc.w\t0\t; '" + f.name + "'");
	  }
	  else if (f.size == 4) {
	    asmout.println ("\tdc.l\t0\t; '" + f.name + "'");
	  }
	  else if (f.size == 8) {
	    asmout.println ("\tdc.l\t0\t; '" + f.name + "'");
	    asmout.println ("\tdc.l\t0");
	  }
	}
      }
      if (odd) {
	asmout.println ("\talign\t2");
      }
	  
      asmout.println();

      index += len+1;
    }
      
    // write the initializer routine
    asmout.println ("\tcode");
    asmout.println ("init_constant_strings:");

    asmout.println ("\tlea.l\tcstring_block(pc),a0");
    asmout.println ("\tmove.l\ta0,cstring_array_data(a5)");

    asmout.println ("\tlea.l\tcstring_array(a5),a0");
    for (int i=0; i<strings.length; i++) {
      String label = ConstantStrings.getLabel (strings[i]);
      asmout.println ("\tmove.l\ta0," + label + "_value(a5)");
    }

    asmout.println ("\trts");
    asmout.println();

  }

  /**
   * write PalmOS PilotMain code.
   */
  void generatePilotMain ()
  {
    asmout.println();
    asmout.println ("\tcode");
    asmout.println();

    asmout.println ("PilotMain:");
    asmout.println ("cmd     set     12");
    asmout.println ("cmdPBP  set     14");
    asmout.println ("launchFlags     set     18");
    asmout.println ("        move.l  #-1,-(a7)");
    asmout.println ("        link    a6,#0");
    asmout.println ("        move.l  a5,-(a7)");

    // safety measure: exit immediately unless normal launch
    asmout.println ("        tst.w   cmd(a6)");
    asmout.println ("        bne.s   PilotMain_leave");

    asmout.println ("        move.w  launchFlags(a6),d0");
    asmout.println ("        and.w   #sysAppLaunchFlagNewGlobals,d0");
    asmout.println ("        bne.s   PilotMain_have_globals");
    asmout.println ("        systrap MemPtrNew(#DataSize.l)");
    asmout.println ("        move.l  a0,a5");
    asmout.println ("PilotMain_have_globals:");

    // rk: inserted to allow for Debuffer debugging
    if (Jump.Breakpoint) {
      asmout.println ("        trap    #8");
      asmout.println ("PilotMain_nobreak:");
    }
    asmout.println ("        clr.l   FirstObject(a5)");
    asmout.println ("        move.l  a7,StackEnd(a5)");
    asmout.println ("        bsr     init_constant_strings");
    asmout.println ("        bsr     all_clinit");
    asmout.println ("        clr.l   d0");
    asmout.println ("        move.w  cmd(a6),d0");
    asmout.println ("        move.l  d0,-(a7)");
    asmout.println ("        move.l  cmdPBP(a6),-(a7)");
    asmout.println ("        clr.l   d0");
    asmout.println ("        move.w  launchFlags(a6),d0");
    asmout.println ("        move.l  d0,-(a7)");
    asmout.println ("        bsr     " + 
		    Klass.forName(Jump.mainClassName)
		    .findMethod("PilotMain", "(III)I")
		    .shortLabel());
    asmout.println ("        lea     12(a7),a7");
    asmout.println ("        move.w  launchFlags(a6),d0");
    asmout.println ("        and.w   #sysAppLaunchFlagNewGlobals,d0");
    asmout.println ("        bne.s   PilotMain_had_globals");
    asmout.println ("        systrap MemChunkFree(a5.l)");
    asmout.println ("PilotMain_had_globals:");

    asmout.println ("PilotMain_leave:");
    asmout.println ("        move.l  (a7)+,a5");
    asmout.println ("        unlk    a6");
    asmout.println ("        addq.l  #4,a7");
    asmout.println ("        rts");
    asmout.println();
    asmout.println ("        dc.b    $89,'PilotMain',0");
    asmout.println ("        align   2");
    asmout.println();
  }

  /**
   * write a single method.
   */
  void generateMethod (MethodInfo method)
  {
    CodeAttribute code = method.attributes.code;
    String label = method.shortLabel();
    Klass myClass = method.cls;
    ConstantPool pool = myClass.constant_pool;
    int t, u;

    if (method.nativeRef != null) {
      NativeRef nr = (NativeRef) method.nativeRef;
      
      asmout.println ("; native " + method);
      asmout.println (";-------------------------------------------------");
      asmout.println();
      asmout.println (label + ":");
      
      nr.copyTo (asmout);
    }
    else if ((code != null) && (code.code != null)) {
      int args = (method.argStackSize() +
		  ((method.access_flags & ACC_STATIC) != 0 ? 0 : 1));
      ASSERT.check(args <= code.max_locals, "Internal error: More args than locals");
      int localonly = code.max_locals - method.argStackSize();
      boolean[] jumpTargets = code.computeJumpTargets();

      asmout.println ("; " + method);
      asmout.println (";-------------------------------------------------");
      asmout.println();
      asmout.println (label + ":");
      
      if (code.exception_table.length > 0) {
	asmout.println("\tpea\t" + label + "__exceptions(pc)");
      } else {
	asmout.println("\tclr.l\t-(a7)");
      }
      
      asmout.println("\tlink\ta6,#-" + 4*localonly);
      if (localonly > 0) {
	for (int j=0; j<localonly; j++) {
	  asmout.println("\tclr.l\t-" + 4*(j+1) + "(a6)");
	}
      }
      
      // generate bytecode translations
      CodeIterator iterator = new CodeIterator (code.code, pool);
      boolean previousNeeded = true;
      int stackDelta = 0;
      boolean[] neededArray = code.codeNeeded;
      do {
	int op = iterator.opcode;
	int index = iterator.opcodeIndex;
	int next = iterator.nextIndex;
	boolean wide = iterator.wide;

	if ((neededArray == null) || neededArray[index]) {
	  generateStackAdjust (stackDelta);
	  stackDelta = 0;
	}

	// insert a label for this instruction?
	if (iterator.wide && jumpTargets[index-1]) {
	  generateStackAdjust (stackDelta);
	  stackDelta = 0;
	  asmout.println (label + "__" + (index-1) + ":");
	}
	else if (jumpTargets[index]) {
	  generateStackAdjust (stackDelta);
	  stackDelta = 0;
	  asmout.println (label + "__" + (index) + ":");
	}
	asmout.println ("\t\t; " + index + ": " + iterator.disassemble());

	if ((neededArray == null) || neededArray[index]) {
	  
	  switch (op) {
	  case op_nop:
	    break;
	  case op_aconst_null:
	    asmout.println ("\tclr.l\t-(a7)");
	    break;
	  case op_iconst_m1:
	    asmout.println ("\tmoveq.l\t#-1,d0");
	    asmout.println ("\tmove.l\td0,-(a7)");
	    break;
	  case op_iconst_0:
	    asmout.println ("\tclr.l\t-(a7)");
	    break;
	  case op_iconst_1:
	  case op_iconst_2:
	  case op_iconst_3:
	  case op_iconst_4:
	  case op_iconst_5:
	    asmout.println ("\tmoveq.l\t#" + (op-op_iconst_0) + ",d0");
	    asmout.println ("\tmove.l\td0,-(a7)");
	    break;
	  case op_lconst_0:
	    asmout.println ("\tclr.l\t-(a7)");
	    asmout.println ("\tclr.l\t-(a7)");
	    break;
	  case op_lconst_1:
	    asmout.println ("\tmoveq.l\t#1,d0");
	    asmout.println ("\tmove.l\td0,-(a7)");
	    asmout.println ("\tclr.l\t-(a7)");
	    break;
	  case op_fconst_0:
	    asmout.println ("\tclr.l\t-(a7)");
	    break;
	  case op_fconst_1:
	    asmout.println ("\tmove.l\tF_CONST_1,-(a7)");
	    break;
	  case op_fconst_2:
	    asmout.println("\tmove.l\tF_CONST_2,-(a7)");
	    break;
	  case op_dconst_0:
	    asmout.println ("\tclr.l\t-(a7)");
	    asmout.println ("\tclr.l\t-(a7)");
	    break;
	  case op_dconst_1:
	    asmout.println ("\tmove.l\tD_CONST_1L,-(a7)");
	    asmout.println ("\tmove.l\tD_CONST_1H,-(a7)");
	    break;
	  case op_bipush:
	    t = iterator.getSignedInteger (0,1);
	    asmout.println("\tmoveq.l\t#" + t + ",d0");
	    asmout.println("\tmove.l\td0,-(a7)");
	    break;
	  case op_sipush:
	    t = iterator.getSignedInteger (0,2);
	    if (t >= -128 && t < 128) {
	      asmout.println("\tmoveq.l\t#" + t + ",d0");
	      asmout.println("\tmove.l\td0,-(a7)");
	    } else {
	      asmout.println("\tmove.l\t#" + t + ",-(a7)");
	    }
	    break;
	  case op_ldc:
	  case op_ldc_w:
	    {
	      if (op == op_ldc) {
		t = iterator.getUnsignedInteger (0,1);
	      }
	      else {
		t = iterator.getUnsignedInteger (0,2);
	      }
	      String sig = pool.getSignature (t);
	      if (sig.equals ("Ljava/lang/String;")) {
		String cString = pool.getString (t);
		asmout.println("\tlea\t" + 
			       ConstantStrings.getLabel (cString) + 
			       "(a5),a0\t; \"" + 
			       commentFilter (cString) + 
			       "\"");
		asmout.println("\tmove.l\ta0,-(a7)");
	      }
	      else if (sig.equals ("I")) {
		int cInt = pool.getInteger (t);
		asmout.println("\tmove.l\t#" + cInt + ",-(a7)");
	      }
	      else if (sig.equals ("F")) {
		int cInt = pool.getFloatAsIntBits (t);
		asmout.println("\tmove.l\t#$" + 
			       Integer.toHexString (cInt) + 
			       ",-(a7)\t; " + pool.getFloat(t));
	      }
	      else {
		ASSERT.fail ("op_ldc " + t + " not int, float or String in " + method);
	      }
	      break;
	    }
	  case op_ldc2_w:
	    {
	      t = iterator.getUnsignedInteger (0,2);
	      String sig = pool.getSignature (t);
	      if (sig.equals ("J")) {
		long cLong = pool.getLong (t);
		asmout.println("\tmove.l\t#$" + 
			       Long.toHexString (cLong & 0xffffffffL) + 
			       ",-(a7)");
		asmout.println("\tmove.l\t#$" + 
			       Long.toHexString ((cLong >> 32) & 0xffffffffL) + 
			       ",-(a7)\t; " + pool.getLong(t));
	      }
	      else if (sig.equals ("D")) {
		long cLong = pool.getDoubleAsLongBits (t);
		asmout.println("\tmove.l\t#$" + 
			       Long.toHexString (cLong & 0xffffffffL) + 
			       ",-(a7)");
		asmout.println("\tmove.l\t#$" + 
			       Long.toHexString ((cLong >> 32) & 0xffffffffL) + 
			       ",-(a7)\t; " + pool.getDouble(t));
	      }
	      else {
		ASSERT.fail ("op_ldc2_w " + t + " not long or double in " + method);
	      }
	      break;
	    }
	  case op_iload:
	  case op_fload:
	  case op_aload:
	    t = iterator.getUnsignedInteger (0, wide ? 2 : 1);
	    asmout.println ("\tmove.l\t" + 
			    localToOffset (t,args) + "(a6),-(a7)");
	    break;
	  case op_lload:
	  case op_dload:
	    t = iterator.getUnsignedInteger (0, wide ? 2 : 1);
	    asmout.println ("\tmove.l\t" +
			    localToOffset (t, args) + "(a6),-(a7)");
	    asmout.println ("\tmove.l\t" +
			    localToOffset (t+1, args) + "(a6),-(a7)");
	    break;
	  case op_iload_0:
	  case op_iload_1:
	  case op_iload_2:
	  case op_iload_3:
	    asmout.println("\tmove.l\t" + 
			   localToOffset (op - op_iload_0, args) + 
			   "(a6),-(a7)");
	    break;
	  case op_lload_0:
	  case op_lload_1:
	  case op_lload_2:
	  case op_lload_3:
	    asmout.println("\tmove.l\t" + 
			   localToOffset (op - op_lload_0, args) + 
			   "(a6),-(a7)");
	    asmout.println("\tmove.l\t" + 
			   localToOffset (op - op_lload_0 + 1, args) + 
			   "(a6),-(a7)");
	    break;
	  case op_fload_0:
	  case op_fload_1:
	  case op_fload_2:
	  case op_fload_3:
	    asmout.println("\tmove.l\t" + 
			   localToOffset (op - op_fload_0, args) + 
			   "(a6),-(a7)");
	    break;
	  case op_dload_0:
	  case op_dload_1:
	  case op_dload_2:
	  case op_dload_3:
	    asmout.println("\tmove.l\t" + 
			   localToOffset (op - op_dload_0, args) + 
			   "(a6),-(a7)");
	    asmout.println("\tmove.l\t" + 
			   localToOffset (op - op_dload_0 + 1, args) + 
			   "(a6),-(a7)");
	    break;
	  case op_aload_0:
	  case op_aload_1:
	  case op_aload_2:
	  case op_aload_3:
	    asmout.println("\tmove.l\t" + 
			   localToOffset (op - op_aload_0, args) + 
			   "(a6),-(a7)");
	    break;
	    
	  case op_iaload:
	  case op_faload:
	  case op_aaload:
	    asmout.println("\tmove.l\t(a7)+,d0");
	    asmout.println("\tmove.l\t(a7),a0");
	    asmout.println("\tbsr\tarray_check");
	    asmout.println("\tasl.l\t#2,d0");
	    asmout.println("\tmove.l\tArray.data(a0),a1");
	    asmout.println("\tmove.l\t0(a1,d0.l),(a7)");
	    break;
	  case op_laload:
	  case op_daload:
	    asmout.println("\tmove.l\t(a7)+,d0");
	    asmout.println("\tmove.l\t(a7),a0");
	    asmout.println("\tbsr\tarray_check");
	    asmout.println("\tasl.l\t#3,d0");
	    asmout.println("\tmove.l\tArray.data(a0),a1");
	    asmout.println("\tmove.l\t4(a1,d0.l),(a7)");
	    asmout.println("\tmove.l\t0(a1,d0.l),-(a7)");
	    break;
	  case op_baload:
	    asmout.println("\tmove.l\t(a7)+,d0");
	    asmout.println("\tmove.l\t(a7),a0");
	    asmout.println("\tbsr\tarray_check");
	    asmout.println("\tmove.l\tArray.data(a0),a1");
	    asmout.println("\tmove.b\t0(a1,d0.l),d1");
	    asmout.println("\text.w\td1");
	    asmout.println("\text.l\td1");
	    asmout.println("\tmove.l\td1,(a7)");
	    break;
	  case op_caload:
	    asmout.println("\tmove.l\t(a7)+,d0");
	    asmout.println("\tmove.l\t(a7),a0");
	    asmout.println("\tbsr\tarray_check");
	    asmout.println("\tmove.l\tArray.data(a0),a1");
	    asmout.println("\tclr.l\td1");
	    asmout.println("\tmove.b\t0(a1,d0.l),d1");
	    asmout.println("\tmove.l\td1,(a7)");
	    break;
	  case op_saload:
	    asmout.println("\tmove.l\t(a7)+,d0");
	    asmout.println("\tmove.l\t(a7),a0");
	    asmout.println("\tbsr\tarray_check");
	    asmout.println("\tasl.l\t#1,d0");
	    asmout.println("\tmove.l\tArray.data(a0),a1");
	    asmout.println("\tmove.w\t0(a1,d0.l),d1");
	    asmout.println("\text.l\td1");
	    asmout.println("\tmove.l\td1,(a7)");
	    break;
	    
	  case op_istore:
	  case op_fstore:
	  case op_astore:
	    t = iterator.getUnsignedInteger (0, wide ? 2 : 1);
	    asmout.println ("\tmove.l\t(a7)+," + 
			    localToOffset (t, args) + "(a6)");
	    break;
	  case op_lstore:
	  case op_dstore:
	    t = iterator.getUnsignedInteger (0, wide ? 2 : 1);
	    asmout.println ("\tmove.l\t(a7)+," + 
			    localToOffset (t+1, args) + "(a6)");
	    asmout.println ("\tmove.l\t(a7)+," + 
			    localToOffset (t, args) + "(a6)");
	    break;
	  case op_istore_0:
	  case op_istore_1:
	  case op_istore_2:
	  case op_istore_3:
	    asmout.println ("\tmove.l\t(a7)+," + 
			    localToOffset (op - op_istore_0, args) + "(a6)");
	    break;
	  case op_lstore_0:
	  case op_lstore_1:
	  case op_lstore_2:
	  case op_lstore_3:
	    asmout.println ("\tmove.l\t(a7)+," + 
			    localToOffset (op - op_lstore_0 + 1, args) + "(a6)");
	    asmout.println ("\tmove.l\t(a7)+," + 
			    localToOffset (op - op_lstore_0, args) + "(a6)");
	    break;
	  case op_fstore_0:
	  case op_fstore_1:
	  case op_fstore_2:
	  case op_fstore_3:
	    asmout.println ("\tmove.l\t(a7)+," + 
			    localToOffset (op - op_fstore_0, args) + "(a6)");
	    break;
	  case op_dstore_0:
	  case op_dstore_1:
	  case op_dstore_2:
	  case op_dstore_3:
	    asmout.println ("\tmove.l\t(a7)+," + 
			    localToOffset (op - op_dstore_0 + 1, args) + "(a6)");
	    asmout.println ("\tmove.l\t(a7)+," + 
			    localToOffset (op - op_dstore_0, args) + "(a6)");
	    break;
	  case op_astore_0:
	  case op_astore_1:
	  case op_astore_2:
	  case op_astore_3:
	    asmout.println ("\tmove.l\t(a7)+," + 
			    localToOffset (op - op_astore_0, args) + "(a6)");
	    break;
	    
	  case op_iastore:
	  case op_fastore:
	    asmout.println("\tmove.l\t(a7)+,d2");
	    asmout.println("\tmove.l\t(a7)+,d0");
	    asmout.println("\tmove.l\t(a7)+,a0");
	    asmout.println("\tbsr\tarray_check");
	    asmout.println("\tasl.l\t#2,d0");
	    asmout.println("\tmove.l\tArray.data(a0),a1");
	    asmout.println("\tmove.l\td2,0(a1,d0.l)");
	    break;
	  case op_aastore:
	    asmout.println("\tmove.l\t(a7)+,a1");
	    asmout.println("\tmove.l\t(a7)+,d0");
	    asmout.println("\tmove.l\t(a7)+,a0");
	    asmout.println("\tbsr\taastore_check");
	    asmout.println("\tasl.l\t#2,d0");
	    asmout.println("\tmove.l\tArray.data(a0),a2");
	    asmout.println("\tmove.l\ta1,0(a2,d0.l)");
	    break;
	  case op_lastore:
	  case op_dastore:
	    asmout.println("\tmove.l\t(a7)+,d2");
	    asmout.println("\tmove.l\t(a7)+,d3");
	    asmout.println("\tmove.l\t(a7)+,d0");
	    asmout.println("\tmove.l\t(a7)+,a0");
	    asmout.println("\tbsr\tarray_check");
	    asmout.println("\tasl.l\t#3,d0");
	    asmout.println("\tmove.l\tArray.data(a0),a1");
	    asmout.println("\tmove.l\td2,0(a1,d0.l)");
	    asmout.println("\tmove.l\td3,4(a1,d0.l)");
	    break;
	  case op_bastore:
	  case op_castore:
	    asmout.println("\tmove.l\t(a7)+,d2");
	    asmout.println("\tmove.l\t(a7)+,d0");
	    asmout.println("\tmove.l\t(a7)+,a0");
	    asmout.println("\tbsr\tarray_check");
	    asmout.println("\tmove.l\tArray.data(a0),a1");
	    asmout.println("\tmove.b\td2,0(a1,d0.l)");
	    break;
	  case op_sastore:
	    asmout.println("\tmove.l\t(a7)+,d2");
	    asmout.println("\tmove.l\t(a7)+,d0");
	    asmout.println("\tmove.l\t(a7)+,a0");
	    asmout.println("\tbsr\tarray_check");
	    asmout.println("\tasl.l\t#1,d0");
	    asmout.println("\tmove.l\tArray.data(a0),a1");
	    asmout.println("\tmove.w\td2,0(a1,d0.l)");
	    break;
	    
	  case op_pop:
	    asmout.println("\taddq.l\t#4,a7");
	    break;
	  case op_pop2:
	    asmout.println("\taddq.l\t#8,a7");
	    break;
	  case op_dup:
	    asmout.println("\tmove.l\t(a7),-(a7)");
	    break;
	  case op_dup_x1: 
	    asmout.println("\tmove.l\t(a7),-(a7)");
	    asmout.println("\tmove.l\t8(a7),4(a7)");
	    asmout.println("\tmove.l\t(a7),8(a7)");
	    break;
	  case op_dup_x2:
	    asmout.println("\tmove.l\t(a7),-(a7)");
	    asmout.println("\tmove.l\t8(a7),4(a7)");
	    asmout.println("\tmove.l\t12(a7),8(a7)");
	    asmout.println("\tmove.l\t(a7),12(a7)");
	    break;
	  case op_dup2: {
	    asmout.println("\tmove.l\t4(a7),-(a7)");
	    asmout.println("\tmove.l\t4(a7),-(a7)");
	    break;
	  }
	  case op_dup2_x1:
	    asmout.println("\tmove.l\t4(a7),-(a7)");
	    asmout.println("\tmove.l\t4(a7),-(a7)");
	    asmout.println("\tmove.l\t16(a7),8(a7)");
	    asmout.println("\tmove.l\t(a7),12(a7)");
	    asmout.println("\tmove.l\t4(a7),16(a7)");
	    break;
	  case op_dup2_x2:
	    asmout.println("\tmove.l\t4(a7),-(a7)");
	    asmout.println("\tmove.l\t4(a7),-(a7)");
	    asmout.println("\tmove.l\t16(a7),8(a7)");
	    asmout.println("\tmove.l\t20(a7),12(a7)");
	    asmout.println("\tmove.l\t(a7),16(a7)");
	    asmout.println("\tmove.l\t4(a7),20(a7)");
	    break;
	  case op_swap:
	    asmout.println("\tmove.l\t(a7),d0");
	    asmout.println("\tmove.l\t4(a7),(a7)");
	    asmout.println("\tmove.l\td0,4(a7)");
	    break;
	  case op_iadd:
	    asmout.println("\tmove.l\t(a7)+,d0");
	    asmout.println("\tadd.l\td0,(a7)");
	    break;
	  case op_ladd:
	    asmout.println ("\tlea.l\t8(a7),a0");
	    asmout.println ("\tlea.l\t16(a7),a1");
	    asmout.println ("\tmove.w\t#$4,ccr");
	    asmout.println ("\taddx.l\t-(a0),-(a1)");
	    asmout.println ("\taddx.l\t-(a0),-(a1)");
	    asmout.println ("\tmove.l\ta1,a7");
	    break;
	  case op_fadd:
	    asmout.println("\tbsr\top_fadd");
	    break;
	  case op_dadd:
	    asmout.println("\tbsr\top_dadd");
	    break;
	  case op_isub:
	    asmout.println("\tmove.l\t(a7)+,d0");
	    asmout.println("\tsub.l\td0,(a7)");
	    break;
	  case op_lsub:
	    asmout.println ("\tlea.l\t8(a7),a0");
	    asmout.println ("\tlea.l\t16(a7),a1");
	    asmout.println ("\tmove.w\t#$4,ccr");
	    asmout.println ("\tsubx.l\t-(a0),-(a1)");
	    asmout.println ("\tsubx.l\t-(a0),-(a1)");
	    asmout.println ("\tmove.l\ta1,a7");
	    break;
	  case op_fsub:
	    asmout.println("\tbsr\top_fsub");
	    break;
	  case op_dsub:
	    asmout.println("\tbsr\top_dsub");
	    break;
	  case op_imul:
	    asmout.println("\tbsr\top_imul");
	    break;
	  case op_lmul:
	    asmout.println("\tbsr\top_lmul");
	    break;
	  case op_fmul:
	    asmout.println("\tbsr\top_fmul");
	    break;
	  case op_dmul:
	    asmout.println("\tbsr\top_dmul");
	    break;
	  case op_idiv:
	    asmout.println("\tbsr\top_idiv");
	    break;
	  case op_ldiv:
	    asmout.println("\tbsr\top_ldiv");
	    break;
	  case op_fdiv:
	    asmout.println("\tbsr\top_fdiv");
	    break;
	  case op_ddiv:
	    asmout.println("\tbsr\top_ddiv");
	    break;
	  case op_irem:
	    asmout.println("\tbsr\top_irem");
	    break;
	  case op_lrem:
	    asmout.println("\tbsr\top_lrem");
	    break;
	  case op_frem:
	    asmout.println("\tbsr\top_frem");
	    break;
	  case op_drem:
	    asmout.println("\tbsr\top_drem");
	    break;
	  case op_ineg:
	    asmout.println("\tneg.l\t(a7)");
	    break;
	  case op_lneg:
	    asmout.println("\tneg.l\t4(a7)");
	    asmout.println("\tnegx.l\t(a7)");
	    break;
	  case op_fneg:
	    asmout.println("\tbsr\top_fneg");
	    break;
	  case op_dneg:
	    asmout.println("\tbsr\top_dneg");
	    break;
	  case op_ishl:
	    asmout.println("\tmove.l\t(a7)+,d0");
	    asmout.println("\tmove.l\t(a7),d1");
	    asmout.println("\tlsl.l\td0,d1");
	    asmout.println("\tmove.l\td1,(a7)");
	    break;
	  case op_lshl:
	    asmout.println("\tbsr\top_lshl");
	    break;
	  case op_ishr:
	    asmout.println("\tmove.l\t(a7)+,d0");
	    asmout.println("\tmove.l\t(a7),d1");
	    asmout.println("\tasr.l\td0,d1");
	    asmout.println("\tmove.l\td1,(a7)");
	    break;
	  case op_lshr:
	    asmout.println("\tbsr\top_lshr");
	    break;
	  case op_iushr:
	    asmout.println("\tmove.l\t(a7)+,d0");
	    asmout.println("\tmove.l\t(a7),d1");
	    asmout.println("\tlsr.l\td0,d1");
	    asmout.println("\tmove.l\td1,(a7)");
	    break;
	  case op_lushr:
	    asmout.println("\tbsr\top_lushr");
	    break;
	  case op_iand:
	    asmout.println("\tmove.l\t(a7)+,d0");
	    asmout.println("\tand.l\td0,(a7)");
	    break;
	  case op_land:
	    asmout.println("\tmove.l\t(a7)+,d0");
	    asmout.println("\tand.l\td0,4(a7)");
	    asmout.println("\tmove.l\t(a7)+,d0");
	    asmout.println("\tand.l\td0,4(a7)");
	    break;
	  case op_ior:
	    asmout.println("\tmove.l\t(a7)+,d0");
	    asmout.println("\tor.l\td0,(a7)");
	    break;
	  case op_lor:
	    asmout.println("\tmove.l\t(a7)+,d0");
	    asmout.println("\tor.l\td0,4(a7)");
	    asmout.println("\tmove.l\t(a7)+,d0");
	    asmout.println("\tor.l\td0,4(a7)");
	    break;
	  case op_ixor:
	    asmout.println("\tmove.l\t(a7)+,d0");
	    asmout.println("\teor.l\td0,(a7)");
	    break;
	  case op_lxor:
	    asmout.println("\tmove.l\t(a7)+,d0");
	    asmout.println("\teor.l\td0,4(a7)");
	    asmout.println("\tmove.l\t(a7)+,d0");
	    asmout.println("\teor.l\td0,4(a7)");
	    break;

	  case op_iinc:
	    if (wide) {
	      t = iterator.getUnsignedInteger (0, 2);
	      u = iterator.getSignedInteger (2, 2);
	    } else {
	      t = iterator.getUnsignedInteger (0, 1);
	      u = iterator.getSignedInteger (1, 1);
	    }
	    asmout.println("\tadd.l\t#" + u + "," + 
			   localToOffset (t, args) + "(a6)");
	    break;
	  case op_i2l:
	    asmout.println("\ttst.l\t(a7)");
	    asmout.println("\tsmi.b\td0");
	    asmout.println("\text.w\td0");
	    asmout.println("\text.l\td0");
	    asmout.println("\tmove.l\td0,-(a7)");
	    break;
	  case op_i2f:
	    asmout.println("\tbsr\top_i2f");
	    break;
	  case op_i2d:
	    asmout.println("\tbsr\top_i2d");
	    break;
	  case op_l2i:
	    asmout.println("\taddq.l\t#4,a7");
	    break;
	  case op_l2f:
	    asmout.println("\tbsr\top_l2f");
	    break;
	  case op_l2d:
	    asmout.println("\tbsr\top_l2d");
	    break;
	  case op_f2i:
	    asmout.println("\tbsr\top_f2i");
	    break;
	  case op_f2l:
	    asmout.println("\tbsr\top_f2l");
	    break;
	  case op_f2d:
	    asmout.println("\tbsr\top_f2d");
	    break;
	  case op_d2i:
	    asmout.println("\tbsr\top_d2i");
	    break;
	  case op_d2l:
	    asmout.println("\tbsr\top_d2l");
	    break;
	  case op_d2f:
	    asmout.println("\tbsr\top_d2f");
	    break;
	  case op_i2b:
	    asmout.println("\tmove.l\t(a7),d0");
	    asmout.println("\text.w\td0");
	    asmout.println("\text.l\td0");
	    asmout.println("\tmove.l\td0,(a7)");
	    break;
	  case op_i2c:
	    asmout.println("\tclr.l\td0");
	    asmout.println("\tmove.b\t3(a7),d0");
	    asmout.println("\tmove.l\td0,(a7)");
	    break;
	  case op_i2s:
	    asmout.println("\tmove.l\t(a7),d0");
	    asmout.println("\text.l\td0");
	    asmout.println("\tmove.l\td0,(a7)");
	    break;
	  case op_lcmp:
	    asmout.println("\tmove.l\t(a7)+,d0");
	    asmout.println("\tmove.l\t(a7)+,d1");
	    asmout.println("\tmove.l\t(a7)+,d2");
	    asmout.println("\tmove.l\t(a7)+,d3");
	    asmout.println("\tsub.l\td1,d3");
	    asmout.println("\tsubx.l\td0,d2");
	    asmout.println("\tsne.b\td0");
	    asmout.println("\tblt.s\t" + label + "__" + index + "a");
	    asmout.println("\tneg.b\td0");
	    asmout.println(label + "__" + index + "a:");
	    asmout.println("\text.w\td0");
	    asmout.println("\text.l\td0");
	    asmout.println("\tmove.l\td0,-(a7)");
	    break;
	  case op_fcmpl:
	    asmout.println("\tbsr\top_fcmpl");
	    break;
	  case op_fcmpg:
	    asmout.println("\tbsr\top_fcmpg");
	    break;
	  case op_dcmpl:
	    asmout.println("\tbsr\top_dcmpl");
	    break;
	  case op_dcmpg:
	    asmout.println("\tbsr\top_dcmpg");
	    break;
	  case op_ifeq:
	    t = iterator.getSignedInteger (0, 2);
	    asmout.println("\tmove.l\t(a7)+,d0");
	    asmout.println("\tbeq\t" + label + "__" + (index + t));
	    break;
	  case op_ifne:
	    t = iterator.getSignedInteger (0, 2);
	    asmout.println("\tmove.l\t(a7)+,d0");
	    asmout.println("\tbne\t" + label + "__" + (index + t));
	    break;
	  case op_iflt:
	    t = iterator.getSignedInteger (0, 2);
	    asmout.println("\tmove.l\t(a7)+,d0");
	    asmout.println("\tbmi\t" + label + "__" + (index + t));
	    break;
	  case op_ifge:
	    t = iterator.getSignedInteger (0, 2);
	    asmout.println("\tmove.l\t(a7)+,d0");
	    asmout.println("\tbpl\t" + label + "__" + (index + t));
	    break;
	  case op_ifgt:
	    t = iterator.getSignedInteger (0, 2);
	    asmout.println("\tmove.l\t(a7)+,d0");
	    asmout.println("\tbgt\t" + label + "__" + (index + t));
	    break;
	  case op_ifle:
	    t = iterator.getSignedInteger (0, 2);
	    asmout.println("\tmove.l\t(a7)+,d0");
	    asmout.println("\tble\t" + label + "__" + (index + t));
	    break;
	  case op_if_icmpeq:
	    t = iterator.getSignedInteger (0, 2);
	    asmout.println("\tmove.l\t(a7)+,d0");
	    asmout.println("\tcmp.l\t(a7)+,d0");
	    asmout.println("\tbeq\t" + label + "__" + (index + t));
	    break;
	  case op_if_icmpne:
	    t = iterator.getSignedInteger (0, 2);
	    asmout.println("\tmove.l\t(a7)+,d0");
	    asmout.println("\tcmp.l\t(a7)+,d0");
	    asmout.println("\tbne\t" + label + "__" + (index + t));
	    break;
	  case op_if_icmplt:
	    t = iterator.getSignedInteger (0, 2);
	    asmout.println("\tmove.l\t(a7)+,d0");
	    asmout.println("\tcmp.l\t(a7)+,d0");
	    asmout.println("\tbgt\t" + label + "__" + (index + t));
	    break;
	  case op_if_icmpge:
	    t = iterator.getSignedInteger (0, 2);
	    asmout.println("\tmove.l\t(a7)+,d0");
	    asmout.println("\tcmp.l\t(a7)+,d0");
	    asmout.println("\tble\t" + label + "__" + (index + t));
	    break;
	  case op_if_icmpgt:
	    t = iterator.getSignedInteger (0, 2);
	    asmout.println("\tmove.l\t(a7)+,d0");
	    asmout.println("\tcmp.l\t(a7)+,d0");
	    asmout.println("\tblt\t" + label + "__" + (index + t));
	    break;
	  case op_if_icmple:
	    t = iterator.getSignedInteger (0, 2);
	    asmout.println("\tmove.l\t(a7)+,d0");
	    asmout.println("\tcmp.l\t(a7)+,d0");
	    asmout.println("\tbge\t" + label + "__" + (index + t));
	    break;
	  case op_if_acmpeq:
	    t = iterator.getSignedInteger (0, 2);
	    asmout.println("\tmove.l\t(a7)+,a0");
	    asmout.println("\tcmpa.l\t(a7)+,a0");
	    asmout.println("\tbeq\t" + label + "__" + (index + t));
	    break;
	  case op_if_acmpne:
	    t = iterator.getSignedInteger (0, 2);
	    asmout.println("\tmove.l\t(a7)+,a0");
	    asmout.println("\tcmpa.l\t(a7)+,a0");
	    asmout.println("\tbne\t" + label + "__" + (index + t));
	    break;
	  case op_goto:
	    t = iterator.getSignedInteger (0, 2);
	    asmout.println("\tbra\t" + label + "__" + (index + t));
	    break;
	  case op_jsr:
	    t = iterator.getSignedInteger (0, 2);
	    asmout.println("\tbsr\t" + label + "__" + (index + t));
	    break;
	  case op_ret:
	    t = iterator.getUnsignedInteger (0, wide ? 2 : 1);
	    asmout.println("\tmove.l\t" + localToOffset (t, args) + "(a6),a0");
	    asmout.println("\tjmp\t(a0)");
	    break;
	  case op_tableswitch: {
	    int def = iterator.getSignedInteger (0, 4);
	    int low = iterator.getSignedInteger (4, 4);
	    int high = iterator.getSignedInteger (8, 4);
	    int off, x;
	    asmout.println("\tmove.l\t(a7)+,d0");
	    for (x=low, off=12; x<=high; x++, off+=4) {
	      asmout.println("\tcmp.l\t#" + x + ",d0");
	      t = iterator.getSignedInteger (off, 4);
	      asmout.println("\tbeq\t" + label + "__" + (index + t));
	    }
	    asmout.println("\tbra\t" + label + "__" + (index + def));
	    break;
	  }
	  case op_lookupswitch: {
	    int def = iterator.getSignedInteger (0, 4);
	    int n = iterator.getSignedInteger (4, 4);
	    int off, x;
	    asmout.println("\tmove.l\t(a7)+,d0");
	    for (x=0, off=8; x<n; x++, off+=8) {
	      t = iterator.getSignedInteger (off, 4);
	      asmout.println("\tcmp.l\t#" + t + ",d0");
	      t = iterator.getSignedInteger (off+4, 4);
	      asmout.println("\tbeq\t" + label + "__" + (index + t));
	    }
	    asmout.println("\tbra\t" + label + "__" + (index + def));
	    break;
	  }
	  case op_ireturn:
	  case op_lreturn:
	  case op_freturn:
	  case op_dreturn:
	  case op_areturn:
	  case op_return:
	    if (next < code.code.length) {
	      asmout.println("\tbra\t" + label + "__out");
	    }
	    break;

	    // TODO? re-implement constant-value optimization?
	    // NO: I don't think it's useful with today's compilers
	  case op_getstatic: {
	    t = iterator.getUnsignedInteger (0, 2);
	    FieldInfo field = pool.getFieldInfo (t);
	    switch (field.signature.charAt(0)) {
	    case 'B':
	    case 'Z':
	      asmout.println("\tmove.b\t" + field.shortLabel() + 
			     "(a5),d0\t; " + field);
	      asmout.println("\text.w\td0");
	      asmout.println("\text.l\td0");
	      asmout.println("\tmove.l\td0,-(a7)");
	      break;
	    case 'C':
	      asmout.println("\tclr.l\td0");
	      asmout.println("\tmove.b\t" + field.shortLabel() + 
			     "(a5),d0\t; " + field);
	      asmout.println("\tmove.l\td0,-(a7)");
	      break;
	    case 'S':
	      asmout.println("\tmove.w\t" + field.shortLabel() + 
			     "(a5),d0\t; " + field);
	      asmout.println("\text.l\td0");
	      asmout.println("\tmove.l\td0,-(a7)");
	      break;
	    case 'F':
	    case 'I':
	    case 'L':
	    case '[':
	      asmout.println("\tmove.l\t" + field.shortLabel() + 
			     "(a5),-(a7)\t; " + field);
	      break;
	    case 'D':
	    case 'J':
	      asmout.println("\tmove.l\t" + field.shortLabel() + 
			     "+4(a5),-(a7)");
	      asmout.println("\tmove.l\t" + field.shortLabel() + 
			     "(a5),-(a7)\t; " + field);
	      break;
	    default:
	      ASSERT.fail("Internal error: Unknown field signature (" + 
			  field + "): " + field.signature);
	      break;
	    }
	    break;
	  }
	  case op_putstatic: {
	    t = iterator.getUnsignedInteger (0, 2);
	    FieldInfo field = pool.getFieldInfo (t);
	    boolean writeable = (field.offset >= 0);
	    switch (field.signature.charAt(0)) {
	    case 'B':
	    case 'C':
	    case 'Z':
	      asmout.println("\tmove.l\t(a7)+,d0");
	      if (writeable) {
		asmout.println("\tmove.b\td0," + field.shortLabel() + 
			       "(a5)\t; " + field);
	      }
	      break;
	    case 'S':
	      asmout.println("\tmove.l\t(a7)+,d0");
	      if (writeable) {
		asmout.println("\tmove.w\td0," + field.shortLabel() + 
			       "(a5)\t; " + field);
	      }
	      break;
	    case 'F':
	    case 'I':
	    case 'L':
	    case '[':
	      if (writeable) {
		asmout.println("\tmove.l\t(a7)+," + field.shortLabel() + 
			       "(a5)\t; " + field);
	      }
	      else {
		asmout.println("\tlea.l\t4(a7),a7");
	      }
	      break;
	    case 'D':
	    case 'J':
	      if (writeable) {
		asmout.println("\tmove.l\t(a7)+," + field.shortLabel() + 
			       "(a5)");
		asmout.println("\tmove.l\t(a7)+," + field.shortLabel() + 
			       "+4(a5)\t; " + field);
	      }
	      else {
		asmout.println("\tlea.l\t8(a7),a7");
	      }
	      break;
	    default:
	      ASSERT.fail("Internal error: Unknown field signature (" + 
			  field + "): " + field.signature);
	      break;
	    }
	    break;
	  }
	  case op_getfield: {
	    t = iterator.getUnsignedInteger (0, 2);
	    FieldInfo field = pool.getFieldInfo (t);
	    asmout.println("\tmove.l\t(a7)+,a0");
	    asmout.println("\tbsr\tnonnull_check");
	    switch (field.signature.charAt(0)) {
	    case 'B':
	    case 'Z':
	      asmout.println("\tmove.b\t" + field.offset + 
			     "(a0),d0\t; " + field);
	      asmout.println("\text.w\td0");
	      asmout.println("\text.l\td0");
	      asmout.println("\tmove.l\td0,-(a7)");
	      break;
	    case 'C':
	      asmout.println("\tclr.l\td0");
	      asmout.println("\tmove.b\t" + field.offset + 
			     "(a0),d0\t; " + field);
	      asmout.println("\tmove.l\td0,-(a7)");
	      break;
	    case 'S':
	      asmout.println("\tmove.w\t" + field.offset + 
			     "(a0),d0\t; " + field);
	      asmout.println("\text.l\td0");
	      asmout.println("\tmove.l\td0,-(a7)");
	      break;
	    case 'F':
	    case 'I':
	    case 'L':
	    case '[':
	      asmout.println("\tmove.l\t" + field.offset + 
			     "(a0),-(a7)\t; " + field);
	      break;
	    case 'D':
	    case 'J':
	      asmout.println("\tmove.l\t" + (field.offset+4) + 
			     "(a0),-(a7)");
	      asmout.println("\tmove.l\t" + field.offset + 
			     "(a0),-(a7)\t; " + field);
	      break;
	    default:
	      ASSERT.fail("Internal error: Unknown field signature (" + 
			  field + "): " + field.signature);
	      break;
	    }
	    break;
	  }
	  case op_putfield: 
	    {
	      t = iterator.getUnsignedInteger (0, 2);
	      FieldInfo field = pool.getFieldInfo (t);
	      boolean writeable = (field.offset >= 0);
	      switch (field.signature.charAt(0)) {
	      case 'B':
	      case 'C':
	      case 'Z':
		asmout.println("\tmove.l\t(a7)+,d0");
		asmout.println("\tmove.l\t(a7)+,a0");
		asmout.println("\tbsr\tnonnull_check");
		if (writeable) {
		  asmout.println("\tmove.b\td0," + field.offset + 
				 "(a0)\t; " + field);
		}
		break;
	      case 'S':
		asmout.println("\tmove.l\t(a7)+,d0");
		asmout.println("\tmove.l\t(a7)+,a0");
		asmout.println("\tbsr\tnonnull_check");
		if (writeable) {
		  asmout.println("\tmove.w\td0," + field.offset + 
				 "(a0)\t; " + field);
		}
		break;
	      case 'F':
	      case 'I':
	      case 'L':
	      case '[':
		asmout.println("\tmove.l\t(a7)+,d0");
		asmout.println("\tmove.l\t(a7)+,a0");
		asmout.println("\tbsr\tnonnull_check");
		if (writeable) {
		  asmout.println("\tmove.l\td0," + field.offset + 
				 "(a0)\t; " + field);
		}
		break;
	      case 'D':
	      case 'J':
		asmout.println("\tmove.l\t(a7)+,d0");
		asmout.println("\tmove.l\t(a7)+,d1");
		asmout.println("\tmove.l\t(a7)+,a0");
		asmout.println("\tbsr\tnonnull_check");
		if (writeable) {
		  asmout.println("\tmove.l\td0," + field.offset + 
				 "(a0)");
		  asmout.println("\tmove.l\td1," + (field.offset + 4) + 
				 "(a0)\t; " + field);
		}
		break;
	      default:
		ASSERT.fail("Internal error: Unknown field signature (" + 
			    field + "): " + field.signature);
		break;
	      }
	      break;
	    }
	  case op_invokevirtual: 
	  case op_invokeinterface:
	    {
	      t = iterator.getUnsignedInteger (0, 2);
	      MethodInfo mm = pool.getMethodInfo (t);
	      int argStackSize = mm.argStackSize();
	      String ret = mm.returnType();
	      MethodInfo mmEff = mm.getEffectiveMethod();
	      if (mmEff != null) {
		asmout.println ("\tmove.l\t" + 
				(4 * argStackSize) + "(a7),a0");
		asmout.println ("\tbsr\tnonnull_check");
		asmout.println ("\tbsr\t" + mmEff.shortLabel() +
				"\t; " + mmEff);
	      }
	      else {
		// a0 = instance
		asmout.println ("\tmove.l\t" + (4 * argStackSize) + 
				"(a7),a0\t; invokevirtual...");
		asmout.println ("\tbsr\tnonnull_check");
		// a0 = ClassInfo structure
		asmout.println ("\tbsr\tgetclassinfo_a0");
		// a0 = vtable relative address
		asmout.println ("\tmove.l\tClassInfo.Vtable(a0),a0");
		// a0 = vtable absolute address
		asmout.println ("\tlea\t0(pc),a1");
		asmout.println ("\tadd.l\ta1,a0");
		// a0 = method relative address
		asmout.println ("\tmove.l\t" + (4 * mm.vtableIndex) + "(a0),a0");
		// a0 = method absolute address
		asmout.println ("\tlea\t0(pc),a1");
		asmout.println ("\tjsr\t0(a0,a1.l)\t; ..." + mm);
	      }
	      asmout.println ("\tlea\t" + (4 * (argStackSize + 1)) + "(a7),a7");
	      switch (ret.charAt(0)) {
	      case 'B':
	      case 'C':
	      case 'I':
	      case 'F':
	      case 'S':
	      case 'Z':
		asmout.println("\tmove.l\td0,-(a7)");
		break;
	      case 'J':
	      case 'D':
		asmout.println("\tmove.l\td1,-(a7)");
		asmout.println("\tmove.l\td0,-(a7)");
		break;
	      case 'L':
	      case '[':
		asmout.println("\tmove.l\ta0,-(a7)");
		break;
	      case 'V':
		break;
	      default:
		ASSERT.fail("Internal error: Unknown return type: " + ret);
		break;
	      }
	      break;
	    }
	  case op_invokespecial:
	    {
	      t = iterator.getUnsignedInteger (0, 2);
	      MethodInfo m = pool.getMethodInfo(t);
	      // In a special case, the method has to be found in a class
	      // different from the one mentioned in the constant pool.
	      if ((!m.name.equals ("<init>")) &&
		  ((myClass.access_flags & ACC_SUPER) != 0) &&
		  myClass.hasSuperclass (m.cls)) {
		// re-search method beginning in our superclass
		m = myClass.superclass.findMethod (m.name, m.signature);
	      }
	      asmout.println ("\tmove.l\t" + 
			      (4 * m.argStackSize()) + "(a7),a0");
	      asmout.println ("\tbsr\tnonnull_check");
	      asmout.println ("\tbsr\t" + m.shortLabel() + 
			      "\t; " + m);
	      asmout.println ("\tlea\t" + (4 * (m.argStackSize() + 1)) + 
			      "(a7),a7");
	      switch (m.returnType().charAt(0)) {
	      case 'B':
	      case 'C':
	      case 'I':
	      case 'F':
	      case 'S':
	      case 'Z':
		asmout.println("\tmove.l\td0,-(a7)");
		break;
	      case 'J':
	      case 'D':
		asmout.println("\tmove.l\td1,-(a7)");
		asmout.println("\tmove.l\td0,-(a7)");
		break;
	      case 'L':
	      case '[':
		asmout.println("\tmove.l\ta0,-(a7)");
		break;
	      case 'V':
		break;
	      default:
		ASSERT.fail("Internal error: Unknown return type: " + m.returnType());
		break;
	      }
	      break;
	    }
	  case op_invokestatic:
	    {
	      t = iterator.getUnsignedInteger (0, 2);
	      MethodInfo m = pool.getMethodInfo(t);
	      asmout.println ("\tbsr\t" + m.shortLabel() + 
			      "\t; " + m);
	      int ss = m.argStackSize();
	      if (ss != 0) {
		asmout.println ("\tlea\t" + (4 * m.argStackSize()) + "(a7),a7");
	      }
	      switch (m.returnType().charAt(0)) {
	      case 'B':
	      case 'C':
	      case 'I':
	      case 'F':
	      case 'S':
	      case 'Z':
		asmout.println("\tmove.l\td0,-(a7)");
		break;
	      case 'J':
	      case 'D':
		asmout.println("\tmove.l\td1,-(a7)");
		asmout.println("\tmove.l\td0,-(a7)");
		break;
	      case 'L':
	      case '[':
		asmout.println("\tmove.l\ta0,-(a7)");
		break;
	      case 'V':
		break;
	      default:
		ASSERT.fail("Internal error: Unknown return type: " + m.returnType());
		break;
	      }
	      break;
	    }
	  case op_new: 
	    {
	      t = iterator.getUnsignedInteger (0, 2);
	      Klass cls = pool.getKlass (t);
	      asmout.println("\tmove.l\t#" + cls.classIndex + ",d0\t; " + cls);
	      asmout.println("\tbsr\top_new");
	      break;
	    }
	  case op_newarray: 
	    {
	      // TODO: check array class implementation
	      t = iterator.getUnsignedInteger (0, 1);
	      String classname = null;
	      int elsize = 0;
	      switch (t) {
	      case  4: classname = "[Z"; elsize = 1; break;
	      case  5: classname = "[C"; elsize = 1; break;
	      case  6: classname = "[F"; elsize = 4; break;
	      case  7: classname = "[D"; elsize = 8; break;
	      case  8: classname = "[B"; elsize = 1; break;
	      case  9: classname = "[S"; elsize = 2; break;
	      case 10: classname = "[I"; elsize = 4; break;
	      case 11: classname = "[J"; elsize = 8; break;
	      default:
		ASSERT.fail("Internal error: Unknown array type: " + t);
		break;
	      }
	      Klass cls = Klass.forName (classname);
	      asmout.println("\tmove.l\t#" + cls.classIndex + ",d0");
	      asmout.println("\tmove.l\t#" + elsize + ",d1");
	      asmout.println("\tbsr\top_newarray");
	      break;
	    }
	  case op_anewarray: 
	    {
	      // TODO: arrays and Class objects...
	      t = iterator.getUnsignedInteger (0, 2);
	      Klass cls = pool.getKlass (t);
	      Klass acls = Klass.forName ("[" + cls.getSignature());
	      asmout.println("\tmove.l\t#" + acls.classIndex + ",d0");
	      // asmout.println("\tmove.l\t#" + cls.Index + ",d1");
	      asmout.println("\tbsr\top_anewarray");
	      break;
	    }
	  case op_arraylength:
	    asmout.println("\tmove.l\t(a7)+,a0");
	    asmout.println("\tbsr\tnonnull_check");
	    asmout.println("\tmove.w\tArray.length(a0),-(a7)");
	    asmout.println("\tclr.w\t-(a7)");
	    break;
	  case op_athrow:
	    asmout.println("\tbsr\top_athrow");
	    break;
	  case op_checkcast: {
	    t = iterator.getUnsignedInteger (0, 2);
	    Klass cls = pool.getKlass (t);
	    int targetIndex = Klass.instanceofTargetClassList.indexOf (cls);
	    Klass exClass = Klass.forName("java/lang/ClassCastException");
	    String subLabel = label + "__" + index + "ok";

	    if (targetIndex < 0) {
	      ASSERT.fail (cls + " not in Klass.instanceofTargetClassList");
	    }

	    asmout.println("\tmove.l\t(a7),a0");
	    asmout.println("\tcmp.w\t#0,a0");
	    asmout.println("\tbeq\t" + subLabel);
	    asmout.println("\tbsr\tgetclassinfo_a0");
	    asmout.println("\tbtst\t#" + (targetIndex % 8) + 
			   ",ClassInfo.Itable+" +
			   (targetIndex / 8) + "(a0)");
	    asmout.println("\tbne.s\t" + subLabel);
	    asmout.println("\tmove.l\t#" + exClass.classIndex + ",d0");
	    asmout.println("\tbsr\top_new");
	    asmout.println("\tmove.l\t(a7),-(a7)");
	    asmout.println("\tbsr\t" + exClass.findMethod("<init>","()V").shortLabel());
	    asmout.println("\tbsr\top_athrow");
	    asmout.println(subLabel + ":");
	    break;
	  }
	  case op_instanceof: {
	    t = iterator.getUnsignedInteger (0, 2);
	    Klass cls = pool.getKlass (t);
	    int targetIndex = Klass.instanceofTargetClassList.indexOf (cls);
	    String subLabel = label + "__" + index + "a";

	    if (targetIndex < 0) {
	      ASSERT.fail (cls + " not in Klass.instanceofTargetClassList");
	    }

	    asmout.println("\tclr.l\td0");
	    asmout.println("\tmove.l\t(a7)+,a0");
	    asmout.println("\tcmp.w\t#0,a0");
	    asmout.println("\tbeq\t" + subLabel);
	    asmout.println("\tbsr\tgetclassinfo_a0");
	    asmout.println("\tbtst\t#" + (targetIndex % 8) + 
			   ",ClassInfo.Itable+" +
			   (targetIndex / 8) + "(a0)");
	    asmout.println("\tsne.b\td0");
	    asmout.println("\tand.l\t#1,d0");
	    asmout.println(subLabel + ":");
	    asmout.println("\tmove.l\td0,-(a7)");
	    break;
	  }
	  case op_monitorenter:
	    asmout.println("\taddq.l\t#4,a7");
	    // asmout.println("\tbsr\top_monitorenter");
	    break;
	  case op_monitorexit:
	    asmout.println("\taddq.l\t#4,a7");
	    // asmout.println("\tbsr\top_monitorexit");
	    break;
	  case op_multianewarray: 
	    {
	      t = iterator.getUnsignedInteger (0, 2);
	      u = iterator.getUnsignedInteger (2, 1);
	      generateMultianewarray (pool.getKlass(t), u, 4*(u-1), 
				      label + "__" + index);
	      asmout.println("\tlea.l\t" + 4*u + "(a7),a7");
	      asmout.println("\tmove.l\ta0,-(a7)");
	      break;
	    }
	  case op_ifnull:
	    t = iterator.getSignedInteger (0, 2);
	    asmout.println("\tmove.l\t(a7)+,d0");
	    asmout.println("\tbeq\t" + label + "__" + (index + t));
	    break;
	  case op_ifnonnull:
	    t = iterator.getSignedInteger (0, 2);
	    asmout.println("\tmove.l\t(a7)+,d0");
	    asmout.println("\tbne\t" + label + "__" + (index + t));
	    break;
	  case op_goto_w:
	    ASSERT.fail("Unimplemented: Opcode 'goto_w'");
	    break;
	  case op_jsr_w:
	    ASSERT.fail("Unimplemented: Opcode 'jsr_w'");
	    break;
	  case op_breakpoint:
	    asmout.println("\tdc.w\tbreakpoint");
	    break;
	  default:
	    ASSERT.fail("Internal error: Unknown bytecode: "+op);
	    break;
	  }
	}
	else {
	  switch (op) {
	  case op_getstatic:
	    {
	      t = iterator.getUnsignedInteger (0, 2);
	      FieldInfo field = pool.getFieldInfo (t);
	      switch (field.signature.charAt(0)) {
	      case 'D':
	      case 'J':
		stackDelta += 2;
		break;
	      default:
		stackDelta += 1;
		break;
	      }
	      break;
	    }
	  case op_putstatic:
	    {
	      t = iterator.getUnsignedInteger (0, 2);
	      FieldInfo field = pool.getFieldInfo (t);
	      switch (field.signature.charAt(0)) {
	      case 'D':
	      case 'J':
		stackDelta -= 2;
		break;
	      default:
		stackDelta -= 1;
		break;
	      }
	      break;
	    }
	  case op_getfield:
	    {
	      t = iterator.getUnsignedInteger (0, 2);
	      FieldInfo field = pool.getFieldInfo (t);
	      switch (field.signature.charAt(0)) {
	      case 'D':
	      case 'J':
		stackDelta += 1;
		break;
	      default:
		stackDelta += 0;
		break;
	      }
	      break;
	    }
	  case op_putfield:
	    {
	      t = iterator.getUnsignedInteger (0, 2);
	      FieldInfo field = pool.getFieldInfo (t);
	      switch (field.signature.charAt(0)) {
	      case 'D':
	      case 'J':
		stackDelta -= 3;
		break;
	      default:
		stackDelta -= 2;
		break;
	      }
	      break;
	    }
	  case op_invokevirtual:
	  case op_invokespecial:
	  case op_invokeinterface:
	    {
	      t = iterator.getUnsignedInteger (0, 2);
	      MethodInfo mm = pool.getMethodInfo (t);
	      int argStackSize = mm.argStackSize();
	      String ret = mm.returnType();
	      switch (ret.charAt(0)) {
	      case 'D':
	      case 'J':
		stackDelta -= argStackSize + 1 - 2;
		break;
	      case 'V':
		stackDelta -= argStackSize + 1 - 0;
		break;
	      default:
		stackDelta -= argStackSize + 1 - 1;
		break;
	      }
	      break;
	    }
	  case op_invokestatic:
	    {
	      t = iterator.getUnsignedInteger (0, 2);
	      MethodInfo mm = pool.getMethodInfo (t);
	      int argStackSize = mm.argStackSize();
	      String ret = mm.returnType();
	      switch (ret.charAt(0)) {
	      case 'D':
	      case 'J':
		stackDelta -= argStackSize - 2;
		break;
	      case 'V':
		stackDelta -= argStackSize - 0;
		break;
	      default:
		stackDelta -= argStackSize - 1;
		break;
	      }
	      break;
	    }
	  case op_multianewarray:
	    t = iterator.getUnsignedInteger (2, 1);
	    stackDelta -= t - 1;
	    break;
	  default:
	    stackDelta += StackDeltas[op];
	    break;
	  }
	  asmout.println ("\t\t; dStack = " + stackDelta);
	}

      } while (iterator.iterate());

      asmout.println (label + "__out:");
      switch (method.returnType().charAt(0)) {
      case 'B':
      case 'C':
      case 'F':
      case 'I':
      case 'S':
      case 'Z':
	asmout.println("\tmove.l\t(a7)+,d0");
	break;
      case 'D':
      case 'J':
	asmout.println("\tmove.l\t(a7)+,d0");
	asmout.println("\tmove.l\t(a7)+,d1");
	break;
      case 'L':
      case '[':
	asmout.println("\tmove.l\t(a7)+,a0");
	break;
      case 'V':
	break;
      default:
	ASSERT.fail("Internal error: Unknown return type at " + method);
	break;
      }
      asmout.println("\tunlk\ta6");
      asmout.println("\taddq.l\t#4,a7");
      asmout.println("\trts");
      asmout.println();
      if (Jump.DebugSymbols) {
	String symbol = method.shortName();
	asmout.println("\tdc.b\t" + (0x80 | symbol.length()) + 
		       ", '" + symbol + "', 0");
	asmout.println("\talign\t2");
	asmout.println();
      }
      if (code.exception_table.length > 0) {
	
	// this is struct ExceptionTableHeader
	asmout.println (label + "__exceptions:");
	asmout.println ("\tdc.l\t" + label);
	asmout.println ("\tdc.w\t" + 4*localonly);
	asmout.println ("\tdc.w\t" + code.exception_table.length);

	for (int j=0; j<code.exception_table.length; j++) {
	  Klass catchClass = Klass.forName (code.exception_table[j].catch_type_name);
	  int spc = code.exception_table[j].start_pc;
	  int epc = code.exception_table[j].end_pc;
	  int hpc = code.exception_table[j].handler_pc;
	  
	  // this is struct ExceptionTableEntry
	  asmout.println ("\tdc.w\t" + label + "__" + spc + "-" + label);
	  asmout.println ("\tdc.w\t" + label + "__" + epc + "-" + label);
	  asmout.println ("\tdc.w\t" + label + "__" + hpc + "-" + label);
	  asmout.println ("\tdc.w\t" + catchClass.classIndex + 
			  "\t; " + catchClass);
	  asmout.println();
	}
      }
    }
    else {
      System.out.println ("code missing for " + method);
    }
  }

  /**
   * generate code for a 'multianewarray' instruction.
   *
   * The code leaves the array in register a0.
   *
   * @param cls            the array class to be instantiated
   * @param dimensions     the number of dimensions
   * @param offsetToLength the stack-relative offset where we find 
   *                       the length for this dimension
   * @param labelPrefix    prefix to be used for branch labels within
   *                       this 'multianewarray' code block
   */
  void generateMultianewarray (Klass cls, int dimensions, 
			       int offsetToLength, String labelPrefix)
  {
    int index = cls.classIndex;
    int elsize = 4;
    int shift = 2;
    String sig = cls.getSignature();
    
    if      (sig.startsWith("[B")) { elsize = 1; shift = 0; }
    else if (sig.startsWith("[C")) { elsize = 1; shift = 0; }
    else if (sig.startsWith("[D")) { elsize = 8; shift = 3; }
    else if (sig.startsWith("[F")) { elsize = 4; shift = 2; }
    else if (sig.startsWith("[I")) { elsize = 4; shift = 2; }
    else if (sig.startsWith("[J")) { elsize = 8; shift = 3; }
    else if (sig.startsWith("[L")) { elsize = 4; shift = 2; }
    else if (sig.startsWith("[S")) { elsize = 2; shift = 1; }
    else if (sig.startsWith("[Z")) { elsize = 1; shift = 0; }
    else if (sig.startsWith("[[")) { elsize = 4; shift = 2; }

    if (dimensions <= 1) {
      // new instance
      asmout.println ("\tmove.l\t#" + cls.classIndex + ",d0\t; " + cls);
      asmout.println ("\tbsr\top_new");
      
      // fill the fields of the new array instance
      asmout.println ("\tmove.l\t(a7),a2");
      asmout.println ("\tmove.l\t" + (offsetToLength+4) + "(a7),d3");
      asmout.println ("\tbsr\tposarray_check_d3");
      asmout.println ("\tmove.w\td3,Array.length(a2)");
      asmout.println ("\tmove.w\t#" + elsize + ",Array.elsize(a2)");

      // allocate the data area
      if (shift > 0) {
	asmout.println ("\tasl.l\t#" + shift + ",d3");
      }
      asmout.println ("\tsystrap\tMemPtrNew(d3.l)");
      asmout.println ("\tmove.l\ta0,Array.data(a2)");
      asmout.println ("\tsystrap\tMemSet(a0.l,d3.l,#0.b)");
      asmout.println ("\tmove.l\t(a7)+,a0");
    }
    else {
      if (!sig.startsWith("[[")) {
	ASSERT.fail ("Wrong array type in 'multianewarray'");
      }

      // data register to be used for the loop
      String loopReg = "d" + (9-dimensions);
      String loopLabel = labelPrefix + "_dim" + dimensions;
      Klass nextClass = Klass.forName (sig.substring (1));

      // new instance
      asmout.println ("\tmove.l\t#" + cls.classIndex + ",d0\t; " + cls);
      asmout.println ("\tbsr\top_new");
      
      // fill the fields of the new array instance
      asmout.println ("\tmove.l\t(a7),a2");
      asmout.println ("\tmove.l\t" + (offsetToLength+4) + "(a7),d3");
      asmout.println ("\tbsr\tposarray_check_d3");
      asmout.println ("\tmove.w\td3,Array.length(a2)");
      asmout.println ("\tmove.w\t#" + elsize + ",Array.elsize(a2)");

      // prepare the loop, allocate the data area
      asmout.println ("\tmove.l\td3," + loopReg);
      asmout.println ("\tsubq.l\t#1," + loopReg);
      if (shift > 0) {
	asmout.println ("\tasl.l\t#" + shift + ",d3");
      }
      asmout.println ("\tsystrap\tMemPtrNew(d3.l)");
      asmout.println ("\tmove.l\ta0,Array.data(a2)");
      asmout.println ("\tmove.l\ta0,-(a7)");

      // the loop
      asmout.println (loopLabel + ":");
      generateMultianewarray (nextClass, dimensions-1, offsetToLength+4, 
			      labelPrefix);
      asmout.println ("\tmove.l\t(a7),a1");
      asmout.println ("\tmove.l\ta0,(a1)+");
      asmout.println ("\tmove.l\ta1,(a7)");
      asmout.println ("\tdbra\t" + loopReg + "," + loopLabel);
      asmout.println ("\taddq.l\t#4,a7");
      asmout.println ("\tmove.l\t(a7)+,a0");
    }
  }

  /**
   * generate a stack-adjust instruction as a replacement 
   * for unneeded bytecodes.
   *
   * @param delta the stack growth to be simulated (in 4-byte int-units)
   */
  void generateStackAdjust (int delta)
  {
    switch (delta) {
    case 2:
      asmout.println ("\tsubq.l\t#8,a7\t; stack adjust for unneeded instructions");
      break;
    case 1:
      asmout.println ("\tsubq.l\t#4,a7\t; stack adjust for unneeded instructions");
      break;
    case 0:
      break;
    case -1:
      asmout.println ("\taddq.l\t#4,a7\t; stack adjust for unneeded instructions");
      break;
    case -2:
      asmout.println ("\taddq.l\t#8,a7\t; stack adjust for unneeded instructions");
      break;
    default:
      asmout.println ("\tlea.l\t" + (-4 * delta) +
		      "(a7),a7\t; stack adjust for unneeded instructions");
      break;
    }
  }		

  /**
   * write the code that calls the &lt;clinit&gt; methods.
   */
  void generateClinitCode ()
  {
    Vector clinits = Klass.clinitOrdering();

    asmout.println ("all_clinit:");
    for (int i=0; i<clinits.size(); i++) {
      MethodInfo m = (MethodInfo) clinits.elementAt(i);
      asmout.println ("\tbsr\t" + m.shortLabel() + "\t; " + m);
    }
    asmout.println ("\trts");
    asmout.println ();
  }

  /**
   * write the needed subroutines. This is controlled mainly by
   * the JVM opcodes used in the program.
   */
  void generateSubroutines ()
  {
    ((NativeRef) allNatives.get("kernel-routines")).copyTo(asmout);

    boolean[] used = Klass.getAllBytecodesUsed();
    asmout.println (";; --- the JVM bytecode implementations ---");
    for (int i=0; i<OPCODE_NAMES.length; i++) {
      if (used[i]) {
	String label = "op_" + OPCODE_NAMES[i];
	NativeRef ref = (NativeRef) allNatives.get (label);
	
	if (ref != null) {
	  asmout.println (label + ":");
	  ref.copyTo (asmout);
	}
      }
    }

    /* not necessary as long as we don't distinguish between Exceptions
    if (used[op_aaload] || used[op_baload] || 
	used[op_caload] || used[op_daload] || 
	used[op_faload] || used[op_iaload] || 
	used[op_laload] || used[op_saload] ||
	used[op_aastore] || used[op_bastore] || 
	used[op_castore] || used[op_dastore] || 
	used[op_fastore] || used[op_iastore] || 
	used[op_lastore] || used[op_sastore]) {
      asmout.println ("throw_ArrayIndexOutOfBoundsException:");
      asmout.println ("\tbra\tthrow_Any");
      asmout.println ();
    }
    */

  }

  /**
   * write the leadout code. This is the DataSize label.
   */
  void generateLeadout ()
  {
    asmout.println();
    asmout.println("\tdata");
    asmout.println("DataSize\tds.w\t0");
    asmout.println();
  }

  /**
   * write the resources section.
   */
  void generateResources ()
  {
    String resFilename = Jump.mainClassName + ".res";

    copyStream (ClassLoader.getSystemResourceAsStream (resFilename), 
		resFilename);
  }

  /*
   *                  helper methods
   * ====================================================
   */

  /**
   * assert compatibility of Java classes.
   */
  void assertCompatibility()
  {
    ASSERT.check (Jump.CHARSIZE == 1, "For AsmBackEnd, Jump.CHARSIZE must be 1");

    // AsmBackEnd needs a specific layout of the java.lang.String class.
    boolean incompatible = false;
    Klass stringClass = Klass.forName ("java/lang/String");
    FieldInfo[] fields = stringClass.fields;

    // the fields needed for Jump
    FieldInfo vField = stringClass.findField ("value");
    FieldInfo cField = stringClass.findField ("count");
    FieldInfo oField = stringClass.findField ("offset");

    // optional field that doesn't hurt
    FieldInfo hField = stringClass.findField ("hash");
	
    // String class must not contain additional non-static fields
    for (int i=0; i<fields.length; i++) {
      FieldInfo f = fields[i];
      if (f.isNeeded() && ((f.access_flags & ACC_STATIC) == 0)) {
	if (f == vField) {
	}
	else if (f == cField) {
	}
	else if (f == oField) {
	}
	else if (f == hField) {
	}
	else {
	  incompatible = true;
	}
      }
    }
	
    if ((vField == null) || 
	(oField == null) ||	
	(cField == null) || 
	incompatible) {

      System.out.println ();
      System.out.println ("***                JAVA VERSION INCOMPATIBILITY                 ***");
      System.out.println ("***        Use a different Java version, e.g. JDK 1.2.x         ***");
      System.out.println ("***  Please report the following lines to Kleberhoff@aol.com !  ***");
      System.out.println ("***        Please add information on your Java version !        ***");
	  
      if (fields != null) {
	System.out.println();
	System.out.println ("   === java.lang.String fields ===");
	for (int i=0; i<fields.length; i++) {
	  fields[i].report();
	}
      }

      ASSERT.fail ("java.lang.String class incompatible with Jump's AsmBackEnd.");
    } 
  }
  
  /**
   * get an abbreviated, one-line version of a string without
   * any dangerous characters, suitable for comments.
   */
  String commentFilter (String s)
  {
    char[] chars;
    if (s.length() > 30) {
      chars = new char[30];
      s.getChars (0, 30, chars, 0);
    }
    else {
      chars = s.toCharArray();
    }

    for (int i=0; i<chars.length; i++) {
      char c = chars[i];
      if (Character.getNumericValue(c) > 255) {
	chars[i] = '*';
      }
      else if (Character.isWhitespace(c)) {
	chars[i] = ' ';
      }
      else if (Character.isISOControl(c)) {
	chars[i] = '.';
      }
    }
    return new String (chars);
  }

  /**
   * translate a JVM local-var index to a frame-pointer (a6) offset.
   *
   * @param local    the index of the local var (from the bytecode).
   * @param argLimit the first local-index not belonging to an argument.
   */
  int localToOffset (int local, int argLimit)
  {
    if (local < argLimit) {
      return -4 * (local - argLimit - 2);
    } else {
      return -4 * (local - argLimit + 1);
    }
  }

  /*
   *               debugging methods
   * ====================================================
   */

  /** the full name of this class. */
  public String toString()
  {
    return "AsmBackEnd " + asmoutFilename;
  }

  void report()
  {
  }


}
